import os
import h5py
import datetime
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

import plotly.graph_objects as go
import plotly.offline as pyo
import inspy as npy
from plotly.subplots import make_subplots
from scipy.interpolate import griddata
from lmfit import Parameters, fit_report, minimize


PG002_D_SPACING     = 3.355  # Angstroms
NI_LATTICE_CONSTANT = 3.524  # Angstroms

def descend_obj(obj, sep='\t'):
    #Iterate through groups in a HDF5 file and prints the groups and datasets names and datasets attributes
    #return the list of key (namelist) and the data title (col_list)
    #the key name (e.g. /entry1/data ) can be used to get the data by open hdf file f= h5py.File, and using f(key)
    #col_list is the key (e.g. s1, s2)
    namelist = []
    col_list = []
    allmotornames= ['Pt.',  's1', 's2', 'qm', 'qh', 'qk', 'ql', 'ei', 'vei', 'ef', 'en', 'time', 'bm1_counts', 'bm2_counts', 'dummy_motor',
                    'VS_left', 'VS_right', 'ps_right', 'ps_left', 'ps_top', 'ps_bottom', 'pa_right', 'pa_left', 'pa_top', 'pa_bottom', 
                    'm1', 'm2', 'mtilt', 'mtrans',  'sgl', 'sgu', 'stl', 'stu', 'a1', 'a2', 'atrans', 'atilt', 'ahfocus', 'avfocus', 
                    'pghf', 'pgvf', 'cuhf', 'cuvf','sensorValueA','sensorValueB','sensorValueC','sensorValueD','setpoint1','setpoint2']
    if type(obj) in [h5py._hl.group.Group, h5py._hl.files.File]:
        for key in obj.keys():
            #print(key)
            if key in allmotornames:
                #print(sep+'group','+',key,':',obj[key])
                #print(obj[key].name)
                namelist.append(obj[key].name)
                col_list.append(key)
            templist1, templist2 = descend_obj(obj[key],sep=sep+'\t')
            namelist = namelist + templist1
            col_list = col_list + templist2
    elif type(obj) == h5py._hl.dataset.Dataset:
        for key in obj.attrs.keys():
            if key in allmotornames:
                #print(sep+'data:\t','-',key,':',obj.attrs[key])
                pass
    
    return namelist, col_list


def h5dump(path, group='/'):
    #print HDF5 file metadata
    #group: you can give a specific group, defaults to the root group

    with h5py.File(path,'r') as file:
         namelist, col_list = descend_obj(file[group])
    #print(namelist)
    #print(col_list)
    return namelist, col_list


def deg2rad(anglex):
    return anglex*np.pi/180.0

def AnglesToQhkl(m2, s1, s2, a2, sgu, sgl, UBMat):

    d_pg002 = PG002_D_SPACING  #d spacing of pg
    phi     = s2

    #m1rad  = deg2rad(m2/2.0)
    m2rad  = deg2rad(m2)

    #a1rad  = deg2rad(a2/2.0)
    a2rad  = deg2rad(a2)

    s1rad  = deg2rad(s1)
    phirad = deg2rad(phi)
    s2rad  = phirad

    sgu_rad = deg2rad(sgu)
    sgl_rad = deg2rad(sgl)


    ki  =   np.pi/np.sin(m2rad/2)/d_pg002

    kf  =   np.pi/np.sin(a2rad/2)/d_pg002

    q   =   np.sqrt(ki*ki+kf*kf-2*ki*kf*np.cos(phirad))

    theta_rad = np.arctan((ki-kf*np.cos(phirad))/kf/np.sin(phirad))

    omega_rad = s1rad-theta_rad

    MatN = np.matrix([(1.0,         0.0,                    0.0             ), 
                      (0.0,         np.cos(sgu_rad),        -np.sin(sgu_rad)), 
                      (0.0,         np.sin(sgu_rad),        np.cos(sgu_rad))])
    
    
    MatM = np.matrix([( np.cos(sgl_rad),             0.0,          np.sin(sgl_rad)), 
                      (             0.0,             1.0,                      0.0), 
                      (-np.sin(sgl_rad),             0.0,          np.cos(sgl_rad))])




    MatO = np.matrix([( np.cos(omega_rad),             -np.sin(omega_rad),                      0.0), 
                      ( np.sin(omega_rad),              np.cos(omega_rad),                      0.0), 
                      (               0.0,                            0.0,                      1.0)])



    MatTh = np.matrix([( np.cos(theta_rad),             -np.sin(theta_rad),                      0.0), 
                       ( np.sin(theta_rad),              np.cos(theta_rad),                      0.0), 
                       (               0.0,                            0.0,                      1.0)])

    MatTemp = np.matmul(MatN.I, MatM.I)

    MatTemp = np.matmul(MatTemp, MatO.I)

    Qtheta = np.array([[q],
                       [0],
                       [0]])
    
    Qnu   = np.matmul(MatTemp, Qtheta)

    Qhkl  = np.matmul(UBMat.I, Qnu)

    Qhkl  = Qhkl/2/np.pi

    return Qhkl

def AnglesToQhklArray(m2, s1, s2, a2, sgu, sgl, UBMat):
    """
    the parameters of this function are arrays or dataframe
    
    """

    d_pg002 = PG002_D_SPACING  #d spacing of pg
    phi     = s2

    #m1rad  = deg2rad(m2/2.0)
    m2rad  = deg2rad(m2)

    #a1rad  = deg2rad(a2/2.0)
    a2rad  = deg2rad(a2)

    s1rad  = deg2rad(s1)
    phirad = deg2rad(phi)
    s2rad  = phirad

    sgu_rad = deg2rad(sgu)
    sgl_rad = deg2rad(sgl)


    ki  =   np.pi/np.sin(m2rad/2)/d_pg002

    kf  =   np.pi/np.sin(a2rad/2)/d_pg002

    q   =   np.sqrt(ki*ki+kf*kf-2*ki*kf*np.cos(phirad))

    theta_rad = np.arctan((ki-kf*np.cos(phirad))/kf/np.sin(phirad))

    omega_rad = s1rad-theta_rad

    MatN = np.matrix([(1.0,         0.0,                    0.0             ), 
                      (0.0,         np.cos(sgu_rad),        -np.sin(sgu_rad)), 
                      (0.0,         np.sin(sgu_rad),        np.cos(sgu_rad))])
    
    
    MatM = np.matrix([( np.cos(sgl_rad),             0.0,          np.sin(sgl_rad)), 
                      (             0.0,             1.0,                      0.0), 
                      (-np.sin(sgl_rad),             0.0,          np.cos(sgl_rad))])




    MatO = np.matrix([( np.cos(omega_rad),             -np.sin(omega_rad),                      0.0), 
                      ( np.sin(omega_rad),              np.cos(omega_rad),                      0.0), 
                      (               0.0,                            0.0,                      1.0)])



    MatTh = np.matrix([( np.cos(theta_rad),             -np.sin(theta_rad),                      0.0), 
                       ( np.sin(theta_rad),              np.cos(theta_rad),                      0.0), 
                       (               0.0,                            0.0,                      1.0)])

    MatTemp = np.matmul(MatN.I, MatM.I)

    MatTemp = np.matmul(MatTemp, MatO.I)

    Qtheta = np.array([[q],
                       [0],
                       [0]])
    
    Qnu   = np.matmul(MatTemp, Qtheta)

    Qhkl  = np.matmul(UBMat.I, Qnu)

    Qhkl  = Qhkl/2/np.pi

    return Qhkl

def strisfloat(num):
    try:
        float(num)
        return True
    except ValueError:
        return False

def strisint(num):
    try:
        int(num)
        return True
    except ValueError:
        return False

def gaussian_residual(pars, x, data=None):
    """Model a decaying sine wave and subtract data."""
    vals = pars.valuesdict()
    A    = vals['A']
    x0   = vals['x0']
    w    = vals['w']
    bg   = vals['bg']

    model = A/(w*np.sqrt(np.pi/(4*np.log(2)))) * np.exp(-4*np.log(2)*(x-x0)**2/w**2) + bg
    if data is None:
        return model
    return model - data

    
def lorentzian_residual(pars, x, data=None):
    """Model a decaying sine wave and subtract data."""
    vals = pars.valuesdict()
    A    = vals['A']
    x0   = vals['x0']
    w    = vals['w']
    bg   = vals['bg']

    model = A*w/((x-x0)**2+w**2/4)/np.pi/2.0 + bg
    if data is None:
        return model
    return model - data

def gaussian(pars, x):
    """Model a decaying sine wave and subtract data."""
    vals = pars.valuesdict()
    A    = vals['A']
    x0   = vals['x0']
    w    = vals['w']
    #bg   = vals['bg']

    return A/(w*np.sqrt(np.pi/(4*np.log(2)))) * np.exp(-4*np.log(2)*(x-x0)**2/w**2) 

def lorentzian(pars, x, data=None):
    """Model a decaying sine wave and subtract data."""
    vals = pars.valuesdict()
    A    = vals['A']
    x0   = vals['x0']
    w    = vals['w']
    #bg   = vals['bg']

    return A*w/((x-x0)**2+w**2/4)/np.pi/2.0 

def three_peaks_on_slope_residual(pars, x, data=None):
    """Model a decaying sine wave and subtract data."""
    vals = pars.valuesdict()

    #peak 1
    A    = vals['A_1']
    x0   = vals['x0_1']
    w    = vals['w_1']
    GL   = vals['GL_1']

    peak_params  =  Parameters()
    peak_params.add('A',    value  =  A)
    peak_params.add('x0',    value  =  x0)
    peak_params.add('w',   value  =  w)

    model = 0
    if GL == 1:
        model = gaussian(peak_params, x)
    elif GL == 2:
        model = lorentzian(peak_params, x)
    elif GL == 0:
        model = model + 0

    #peak 2
    A    = vals['A_2']
    x0   = vals['x0_2']
    w    = vals['w_2']
    GL   = vals['GL_2']

    peak_params['A'].value   =  A
    peak_params['x0'].value  =  x0
    peak_params['w'].value   =  w

    if GL == 1: #gaussian
        model = model + gaussian(peak_params, x)
    elif GL == 2:
        model = model + lorentzian(peak_params, x)
    elif GL == 0:
        model = model + 0

    #peak 3
    A    = vals['A_3']
    x0   = vals['x0_3']
    w    = vals['w_3']
    GL   = vals['GL_3']

    peak_params['A'].value   =  A
    peak_params['x0'].value  =  x0
    peak_params['w'].value   =  w

    if GL == 1:
        model = model + gaussian(peak_params, x)
    elif GL == 2:
        model = model + lorentzian(peak_params, x)
    elif GL == 0:
        model = model + 0

    #slope background
    bg    = vals['bg']
    slope = vals['slope']

    model = model + slope*x + bg
    
    if data is None:
        return model
    return model - data

def lorenz_residual(pars, x, data=None):
    """Model a lorenz and subtract data."""
    vals = pars.valuesdict()
    A   = vals['A']
    kai = vals['kai']
    x0  = vals['x0']
    bg  = vals['bg']
    
    model = A/(kai**2+(x-x0)**2)    + bg
    if data is None:
        return model
    return model - data

def lorenzsq_residual(pars, x, data=None):
    """Model a decaying lorenzsq and subtract data."""
    vals = pars.valuesdict()
    A   = vals['A']
    kai = vals['kai']
    x0  = vals['x0']
    bg  = vals['bg']
    
    model = A/(kai**2+(x-x0)**2)**2    + bg
    if data is None:
        return model
    return model - data

#this is a comparison of lorentz and lorentzsq
def lorenz_gaussian_residual(pars, x, data=None):
    """Model a decaying lorenz + gaussian and subtract data."""
    vals = pars.valuesdict()
    A    = vals['A']
    kai = vals['kai']
    x0  = vals['x0']
    bg  = vals['bg']
    B   = vals['B']
    w   = vals['w']
    
    model = A/(kai**2+(x-x0)**2)**2    +  B/(w*np.sqrt(np.pi/(4*np.log(2)))) * np.exp(-4*np.log(2)*(x-x0)^2/w^2) + bg
    if data is None:
        return model
    return model - data  

def lorentzian_asym(pars, x, data=None):
    """Model a decaying sine wave and subtract data."""
    vals   = pars.valuesdict()
    A      = vals['A']
    gamma0 = vals['gamma0']
    x0     = vals['x0']
    asym   = vals['asym']
    bg     = vals['bg']
    res_w  = vals['res_w']
    
    step=x[1]-x[0]

    x_range=np.linspace(x0-3*step, x0+3*step, 7)
    res_g  = 1/(res_w*np.sqrt(np.pi/(4*np.log(2)))) * np.exp(-4*np.log(2)*(x_range-x0)**2/res_w**2)

    gg  = 2*gamma0/(1+np.exp(asym*(x-x0)))

    model =  A*gg/(gg**2+(x-x0)**2)-A*gg/(gg**2+(x+x0)**2)

    conv_model = np.convolve(model, res_g, 'same') + bg
    
    
    if data is None:
        return conv_model
    return conv_model - data


def ni_s2_residual(pars, x, data=None):
    """Model a decaying sine wave and subtract data."""
    vals        = pars.valuesdict()
    #m1         = vals['m1']
    s2_offset   = vals['s2_offset']
    wavelen     = vals['wavelen']
    
    ni_hkl      = [[0.5, 0.5, 0.5],
                   [1, 0, 0],
                   [1, 1, 0],
                   [0.5, 0.5, 1.5],
                   [1, 1, 1],
                   [2, 0, 0]]


    ni_a        = NI_LATTICE_CONSTANT
    
    calc_s2=np.array([])
    
    for ii in  range(len(x)):
        s2 = -(np.arcsin(wavelen*np.sqrt(ni_hkl[ii][0]**2 + ni_hkl[ii][1]**2 + ni_hkl[ii][2]**2)/ni_a/2)*360/np.pi+s2_offset)
        calc_s2 = np.append(calc_s2,s2)

    if data is None:
        return calc_s2
    return calc_s2 - data

def ni_s2_residual_30p5meV(pars, x, data=None):
    """Model a decaying sine wave and subtract data."""
    vals = pars.valuesdict()
    #m1    = vals['m1']
    s2_offset   = vals['s2_offset']
    wavelen     = vals['wavelen']
    
    ni_hkl=[[0.5, 0.5, 0.5],
            [1, 0, 0],
            [1, 1, 0],
            [1, 1, 1],
            [2, 0, 0],
            [2, 2, 0]]


    ni_a = NI_LATTICE_CONSTANT
    
    calc_s2=np.array([])
    
    for ii in  range(len(x)):
        s2 = -(np.arcsin(wavelen*np.sqrt(ni_hkl[ii][0]**2 + ni_hkl[ii][1]**2 + ni_hkl[ii][2]**2)/ni_a/2)*360/np.pi+s2_offset)
        calc_s2 = np.append(calc_s2,s2)

    if data is None:
        return calc_s2
    return calc_s2 - data
    

def ni_s2_residual_cu(pars, x, data=None):
    """Model a decaying sine wave and subtract data."""
    vals = pars.valuesdict()
    #m1    = vals['m1']
    s2_offset   = vals['s2_offset']
    wavelen     = vals['wavelen']
    
    ni_hkl=[[1, 1, 1],
            [2, 0, 0],
            [2, 1, 1],[2, 2, 0],[3, 1, 1],[4, 0, 0]]


    ni_a = NI_LATTICE_CONSTANT
    
    calc_s2=np.array([])
    
    for ii in  range(len(x)):
        s2 = -(np.arcsin(wavelen*np.sqrt(ni_hkl[ii][0]**2 + ni_hkl[ii][1]**2 + ni_hkl[ii][2]**2)/ni_a/2)*360/np.pi+s2_offset)
        calc_s2 = np.append(calc_s2,s2)

    if data is None:
        return calc_s2
    return calc_s2 - data
    
def fit_peak(x, data=None, func='G', initial=None):
    
    fitted_data =  pd.DataFrame([])
    fitted_params   =  pd.DataFrame(columns = ['A', 'A_err', 'w','w_err', 'x0', 'x0_err', 'bg', 'bg_err'])
    dataX=x
    dataY=data
    if isinstance(dataY, pd.DataFrame):
        dataX = dataX.to_numpy()
    if isinstance(data, pd.DataFrame):
        dataY = dataY.to_numpy()
    dataX=dataX[np.logical_not(np.isnan(dataX))]
    dataY=dataY[np.logical_not(np.isnan(dataY))]
    #print(dataX)
    #print(dataY)
    peak_params  =  Parameters()
    if initial is None:
        pointnum=len(dataY)

        max_idx = np.argmax(dataY)
        peak_params.add('A',  value = dataY[max_idx])  # peak height
        peak_params.add('w',  value = 4.0*np.abs(dataX[pointnum-1] - dataX[0])/pointnum)  # width set 4 times of stepsize
        peak_params.add('x0', value = dataX[max_idx])  # peak center
        peak_params.add('bg', value = (dataY[0] + dataY[pointnum-1]) / 2)  # background, keep as before

    else: 
        peak_params.add('A',    value  =  initial[0])
        peak_params.add('w',    value  =  initial[1])
        peak_params.add('x0',   value  =  initial[2])
        peak_params.add('bg',   value  =  initial[3])



    if func   == "G":
        out   = minimize(gaussian_residual, peak_params, args=(dataX,),   kws={'data': dataY})
    elif func == "L":
        out   = minimize(lorentzian_residual, peak_params, args=(dataX,), kws={'data': dataY})
    else:
        print("Choose wrong function!")
        return

    xrange      =  np.linspace(dataX[0], dataX[-1], 101)
    fitted_data =  pd.merge(fitted_data, pd.DataFrame(xrange, columns=['X']), how='outer', left_index=True, right_index=True,suffixes=('', ''))

    if   func == "G":
        y_fitted    =  gaussian_residual(out.params, xrange)
    elif func == "L":
        y_fitted    =  lorentzian_residual(out.params, xrange)
    else:
        print("Choose wrong function!")
        return
    
    fitted_data =  pd.merge(fitted_data, pd.DataFrame(y_fitted, columns=['Y_fit']), how='outer', left_index=True, right_index=True,suffixes=('', ''))
    
    new_row ={}
    for name, param in out.params.items():
        new_row[name]        =  [param.value]
        new_row[name+"_err"] =  [param.stderr]

    newpd_row = pd.DataFrame.from_dict(new_row,orient='columns')
    newpd_row = newpd_row.astype(fitted_params.dtypes.to_dict())

    pd.set_option('display.expand_frame_repr', False)
    fitted_params =pd.concat([fitted_params, newpd_row],  join='outer')

    return fitted_params, fitted_data


def angle2(x, y, z, h, k, l, lattice):
    r"""Function necessary for Prefactor functions
    """
    latticestar = npy.tools._star(lattice)[-1]
    #print(latticestar)
    #print( 2 * np.pi * (h * x + k * y + l * z) / npy.tools._modvec([x, y, z], lattice) / npy.tools._modvec([h, k, l], latticestar))

    return np.arccos( 2 * np.pi * (h * x + k * y + l * z) / npy.tools._modvec([x, y, z], lattice) / npy.tools._modvec([h, k, l], latticestar))

def SqwDemo(H, K, L, W, p):
    #Example Scattering function for convolution tests
    
    en1     =    p[0]   #peak 1
    en2     =    p[1]   #peak 2
    IntRatio=    p[2]
    Gamma1 =     p[3]
    Gamma2 =     p[4]  #this is the reason, gamma2 was not used. 
    #Int   =     p[5]
    #Bg    =     p[6]
    temp   =     p[7]
    
    A1=1/((W-en1)**2+Gamma1**2)
    B1=1/((W+en1)**2+Gamma1**2)
    
    A2=1/((W-en2)**2+Gamma2**2)
    B2=1/((W+en2)**2+Gamma2**2)
    
    BF=1+1/(1-np.exp(-W/temp/0.08713))
    
    sqw0=Gamma1*BF*(A1-B1) + IntRatio*Gamma2*BF*(A2-B2)
    sqw1=Gamma1*BF*(A1-B1) + IntRatio*Gamma2*BF*(A2-B2)

    sqw = np.vstack((sqw0, sqw1, sqw1))

    return sqw

def AsymSqwDemo(H,K,L,W,p):
    #J. Phys.: Condens. Matter 9 (1997) 1599–1608. Printed in the UK PII: S0953-8984(97)73830-X
    #An inelastic neutron scattering study of the Kondo semiconductor CeNiSn in high magnetic field
    #S Raymondy, L P Regnaulty, T Satoz, H Kadowakiz, N Pykax,
    #G Nakamotok, T Takabatakek, H Fujiik, Y Isikawa{, G Lapertoty and J Flouquet

    Delta1     =    p[0]   #peak 1
    Delta2     =    p[1]   #peak 2
    ratio      =    p[2]
    Gamma1     =    p[3]
    Gamma2     =    p[4]  #this is the reason, gamma2 was not used. 
    #Int      =     p[5]
    #Bg       =     p[6]
    temp       =    p[7]
    #print(p)

    B1 = 1/np.sqrt(W**2-(complex(Delta1, Gamma1))**2)

    C1 = np.abs(complex(Delta1, Gamma1))

    #D1 = C1*W*B1.real*B1.real          #normalize to the gap energy
    D1 = W*B1.real*B1.real              #without normalizing to the gap energy

    BF= 1+1/(1-np.exp(-W/temp/0.08713))

    B2 = 1/np.sqrt(W**2-(complex(Delta2, Gamma2))**2)

    C2 = np.abs(complex(Delta2, Gamma2))

    #D2 = C2*W*B2.real*B2.real          #normalize to the gap energy
    D2 = W*B2.real*B2.real              #without normalizing to the gap energy

    #print(ratio)

    sqw0 = BF*D1  + ratio*BF*D2

    sqw = np.vstack((sqw0, sqw0, sqw0))

    return sqw



def AsymSqwDynSuscep(H,K,L,W,p):
    #PHYSICAL REVIEW B 102, 014443 (2020)
    # Bose factor (Detailed Balance)  
    # Using 0.08617 for k_B in meV/K
    #I_BG, A_Q, E0, Gamma, temp

    Delta1     =    p[0]   #peak 1  E0
    Delta2     =    p[1]   #peak 2
    ratio      =    p[2]
    Gamma1     =    p[3]
    Gamma2     =    p[4]  #this is the reason, gamma2 was not used. 
    #Int      =     p[5]  #A_Q
    #Bg       =     p[6]
    temp       =    p[7]
    #print(p)

    bose_factor = 1 + 1 / (np.exp(W / (temp * 0.08617)) - 1)
    
    # Complex energy term
    epsilon1_Q = complex(Delta1, Gamma1)
    
    # Complex energy term
    epsilon2_Q = complex(Delta2, Gamma2)
    # The Response Function term
    response1 = 1 / np.sqrt(W**2 - epsilon1_Q**2) 
    response2 = 1 / np.sqrt(W**2 - epsilon2_Q**2)
    
    # Final Formula
    sqw0 =  bose_factor * W * (response1.real**2  + ratio* response2.real**2)

    sqw  = np.vstack((sqw0, sqw0, sqw0))

    return  sqw

def calculate_intensity_aq(E, I_BG, A_Q, E0, Gamma, temp):
    """
    #PHYSICAL REVIEW B 102, 014443 (2020)
    Calculates the intensity I(Q, E) from the phenomenological function:
    I(Q, E) = I_BG + (n(E) + 1) * E * A_Q * [Re(1/sqrt(E^2 - eps^2))]^2

    """
    # Boltzmann constant in meV/K
    kb = 0.08617 
    
    bose_factor = 1 / (1 - np.exp(-E / (kb * temp)))
    
    # 2. Complex energy epsilon_Q
    epsilon_Q = E0 + 1j * Gamma
    
    # 3. The Response Function Term: [Re(1 / sqrt(E^2 - epsilon_Q^2))]^2
    # Note: np.sqrt on a complex array returns the principal square root
    response = 1.0 / np.sqrt(E**2 - epsilon_Q**2)
    response_term = (response.real)**2
    
    # 4. Final Equation Assembly
    intensity = bose_factor * E * response_term
    
    return intensity


def AsymLorentzianSqw(H,K,L,W,p):
    """Model a decaying sine wave and subtract data."""

    Delta1     =    p[0]   #peak 1
    Delta2     =    p[1]   #peak 2
    ratio      =    p[2]
    Gamma1     =    p[3]
    Gamma2     =    p[4]  #this is the reason, gamma2 was not used. 
    #Int   =     p[5]
    #Bg    =     p[6]
    asym       =    p[7]
    
    #step    = x[1]-x[0]
    #x_range = np.linspace(x0-3*step,x0+3*step, 7 )  # x_range is W  
    #res_g   = 1/(res_w*np.sqrt(np.pi/(4*np.log(2)))) * np.exp(-4*np.log(2)*(W-x0)**2/res_w**2)

    gg1  = 2*Gamma1/(1+np.exp(asym*(W-Delta1)))
    gg2  = 2*Gamma2/(1+np.exp(asym*(W-Delta2)))

    int1 =  gg1/(gg1**2+(W-Delta1)**2)-gg1/(gg1**2+(W+Delta1)**2)
    int2 =  gg2/(gg2**2+(W-Delta2)**2)-gg2/(gg2**2+(W+Delta2)**2)
    sqw0= int1 + ratio*int2

    sqw = np.vstack((sqw0, sqw0, sqw0))
    return sqw



def PrefDemo(H, K, L, W, EXP, p):
    #Prefactor example for convolution tests
    
    [sample, rsample] = EXP.get_lattice()

    q2 = npy.tools._modvec([H, K, L], rsample) ** 2
    if len(p) > 8:
        AA =   p[8]
        aa =   p[9]
        BB =   p[10]
        bb =   p[11]
        CC =   p[12]
        cc =   p[13]
        DD =   p[14]

        # Now, use the Jane Brown approximate expression for Co2+
        sd = q2 / (16 * np.pi ** 2)
        ff = AA*np.exp(-aa*sd) + BB*np.exp(-bb*sd) + CC*np.exp(-cc*sd) + DD
    else:
        ff= np.ones_like(q2)

    alphax = angle2(1, 0, 0, H, K, L, sample)
    alphay = angle2(0, 1, 0, H, K, L, sample)
    alphaz = angle2(0, 0, 1, H, K, L, sample)

    polx = np.sin(alphax) ** 2
    poly = np.sin(alphay) ** 2
    polz = np.sin(alphaz) ** 2

    prefactor = np.zeros((3, len(H)))
    prefactor[0, :] = ff ** 2.0 * polx * p[5]
    prefactor[1, :] = ff ** 2.0 * poly * p[5]
    prefactor[2, :] = ff ** 2.0 * polz * p[5]

    bgr = np.ones(H.shape) * p[6]

    return [prefactor, bgr]


#piecewise logistic-decay function definition
def q_x_LogRiseExpDecay(x, A0, x0, c, k):
    """
    Defines the piecewise logistic-decay function Q(x).

    Q(x) = { 2*A0 / (1 + exp(-c*(x - x0)))  if x < x0
           { A0 * exp(-k*(x - x0))         if x >= x0
    
    This function is continuous at x = x0.
    """
    
    # Initialize an array for the function values
    Q = np.zeros_like(x)
    
    # Boolean mask for the logistic (left) side (x < x0)
    mask_logistic = x < x0
    # Boolean mask for the exponential (right) side (x >= x0)
    mask_exponential = x >= x0
    
    # 1. Calculate the Logistic Rise (x < x0)
    # The limit as x -> -inf is 0, and the limit as x -> x0 is A0.
    Q[mask_logistic] = (2 * A0) / (1 + np.exp(-c * (x[mask_logistic] - x0)))
    
    # 2. Calculate the Exponential Decay (x >= x0)
    # The value at x0 is A0, and the limit as x -> +inf is 0.
    Q[mask_exponential] = A0 * np.exp(-k * (x[mask_exponential] - x0))
    
    return Q



def AsymLogREDSqw(H,K,L,W,p):
    #Logistic Rise - Exponential Decay

    e1     =    p[0]   #peak 1 in energy
    e2     =    p[1]   #peak 2 in energy
    ratio  =    p[2]
    c1     =    p[3]   
    k1     =    p[4] 
    c2     =    p[5]   #this is the reason, gamma2 was not used. 
    k2     =    p[6] 
    #Int   =    p[7]
    #Bg    =    p[8]
    temp   =    p[9]
    #print(p)
    Q1 = q_x_LogRiseExpDecay(W, 1, e1, c1, k1)   #amplitude set to 1 because intensity will be in prefactor
    Q2 = q_x_LogRiseExpDecay(W, 1, e2, c2, k2)   #amplitude set to 1 because intensity will be in prefactor

    BF= 1+1/(1-np.exp(-W/temp/0.08713))

    sqw0 = BF*Q1  + ratio*BF*Q2

    sqw = np.vstack((sqw0, sqw0, sqw0))

    return sqw



def PrefLogRED(H, K, L, W, EXP, p):
    #Prefactor example for convolution tests

    Azero =    p[7]
    Bg    =    p[8]
    
    [sample, rsample] = EXP.get_lattice()

    q2 = npy.tools._modvec([H, K, L], rsample) ** 2
    if len(p) > 10:
        AA =   p[10]
        aa =   p[11]
        BB =   p[12]
        bb =   p[13]
        CC =   p[14]
        cc =   p[15]
        DD =   p[16]

        # Now, use the Jane Brown approximate expression for Co2+
        sd = q2 / (16 * np.pi ** 2)
        ff = AA*np.exp(-aa*sd) + BB*np.exp(-bb*sd) + CC*np.exp(-cc*sd) + DD
    else:
        ff= np.ones_like(q2)

    alphax = angle2(1, 0, 0, H, K, L, sample)
    alphay = angle2(0, 1, 0, H, K, L, sample)
    alphaz = angle2(0, 0, 1, H, K, L, sample)

    polx = np.sin(alphax) ** 2
    poly = np.sin(alphay) ** 2
    polz = np.sin(alphaz) ** 2

    prefactor = np.zeros((3, len(H)))
    prefactor[0, :] = ff ** 2.0 * polx * Azero
    prefactor[1, :] = ff ** 2.0 * poly * Azero
    prefactor[2, :] = ff ** 2.0 * polz * Azero

    bgr = np.ones(H.shape) * Bg

    return [prefactor, bgr]


def q_x_LorenRiseExpDecay(x_array, A0, x0, gamma, k):
    """
    Calculates the stitched Lorentzian-Exponential excitation probability.
    
    Parameters:
    x_array (ndarray): Array of incident energy values.
    A0 (float): Peak amplitude at x0.
    x0 (float): Threshold/Resonant energy (Delta E).
    gamma (float): Lorentzian Half-Width at Half-Maximum (HWHM).
    k (float): Exponential decay constant for x >= x0.
    
    Returns:
    ndarray: Probability Q(x) for each value in x_array.
    """
    # Define the Lorentzian rise for x < x0
    lorentzian = A0 / (1 + ((x_array - x0) / gamma)**2)
    
    # Define the Exponential decay for x >= x0
    exponential = A0 * np.exp(-k * (x_array - x0))
    
    # Stitch them together based on the condition x < x0
    return np.where(x_array < x0, lorentzian, exponential)

#Lorentzian-decay function definition
def AsymLorenREDSqw(H,K,L,W,p):
    #Logistic Rise - Exponential Decay

    e1     =    p[0]   #peak 1 in energy
    e2     =    p[1]   #peak 2 in energy
    ratio  =    p[2]
    g1     =    p[3]   
    k1     =    p[4] 
    g2     =    p[5]  #this is the reason, gamma2 was not used. 
    k2     =    p[6] 
    #Int   =    p[7]
    #Bg    =    p[8]
    temp   =    p[9]
    #print(p)
    Q1 = q_x_LorenRiseExpDecay(W, 1, e1, g1, k1)
    Q2 = q_x_LorenRiseExpDecay(W, 1, e2, g2, k2)


    BF= 1+1/(1-np.exp(-W/temp/0.08713))
    #print(ratio)

    sqw0 = BF*Q1  + ratio*BF*Q2

    sqw = np.vstack((sqw0, sqw0, sqw0))

    return sqw



def PrefLorenRED(H, K, L, W, EXP, p):
    #Prefactor example for convolution tests

    Azero =    p[7]
    Bg    =    p[8]
    
    [sample, rsample] = EXP.get_lattice()

    q2 = npy.tools._modvec([H, K, L], rsample) ** 2
    if len(p) > 10:
        AA =   p[10]
        aa =   p[11]
        BB =   p[12]
        bb =   p[13]
        CC =   p[14]
        cc =   p[15]
        DD =   p[16]

        # Now, use the Jane Brown approximate expression for Co2+
        sd = q2 / (16 * np.pi ** 2)
        ff = AA*np.exp(-aa*sd) + BB*np.exp(-bb*sd) + CC*np.exp(-cc*sd) + DD
    else:
        ff= np.ones_like(q2)

    alphax = angle2(1, 0, 0, H, K, L, sample)
    alphay = angle2(0, 1, 0, H, K, L, sample)
    alphaz = angle2(0, 0, 1, H, K, L, sample)

    polx = np.sin(alphax) ** 2
    poly = np.sin(alphay) ** 2
    polz = np.sin(alphaz) ** 2

    prefactor = np.zeros((3, len(H)))
    prefactor[0, :] = ff ** 2.0 * polx * Azero
    prefactor[1, :] = ff ** 2.0 * poly * Azero
    prefactor[2, :] = ff ** 2.0 * polz * Azero

    bgr = np.ones(H.shape) * Bg

    return [prefactor, bgr]



def PrefPhononDemo(H, K, L, W, EXP, p):
    #Prefactor example for convolution tests
    
    [sample, rsample] = EXP.get_lattice()

    alphax = angle2(1, 0, 0, H, K, L, sample)
    alphay = angle2(0, 1, 0, H, K, L, sample)
    alphaz = angle2(0, 0, 1, H, K, L, sample)

    polx = np.sin(alphax) ** 2
    poly = np.sin(alphay) ** 2
    polz = np.sin(alphaz) ** 2

    prefactor = np.zeros((3, len(H)))
    prefactor[0, :] = polx * p[5]
    prefactor[1, :] = poly * p[5]
    prefactor[2, :] = polz * p[5]

    bgr = np.ones(H.shape) * p[6]

    return [prefactor, bgr]

def SqwDemoQscan(H, K, L, W, p):
    # for the q scan
    # Extract the three parameters contained in "p":
    q1    =   p[0]                    # peak1 q position
    q2    =   p[1]                    # peak1 q position
    ratio =   p[2]                    # Intensity ratio 
    w1    =   p[3] 
    w2    =   p[4]                    # peak width

    A1=np.zeros(W.shape)
    
    A2=np.zeros(W.shape)

    if np.abs(H[0]-H[-1])/len(H) > 0.0005:
        A1=1/(w1*np.sqrt(np.pi/(4*np.log(2)))) * np.exp(-4*np.log(2)*(H-q1)**2/w1**2)
        A2=1/(w2*np.sqrt(np.pi/(4*np.log(2)))) * np.exp(-4*np.log(2)*(H-q2)**2/w2**2)
        #print("H")     
    if np.abs(K[0]-K[-1])/len(K) > 0.0005:
        A1=1/(w1*np.sqrt(np.pi/(4*np.log(2)))) * np.exp(-4*np.log(2)*(K-q1)**2/w1**2)
        A2=1/(w2*np.sqrt(np.pi/(4*np.log(2)))) * np.exp(-4*np.log(2)*(K-q2)**2/w2**2)
        #print("K")     
    if np.abs(L[0]-L[-1])/len(L) > 0.0005:
        A1=1/(w1*np.sqrt(np.pi/(4*np.log(2)))) * np.exp(-4*np.log(2)*(L-q1)**2/w1**2)
        A2=1/(w2*np.sqrt(np.pi/(4*np.log(2)))) * np.exp(-4*np.log(2)*(L-q2)**2/w2**2)   
        #print("L")             

    sqw0 = A1 + ratio*A2
    sqw1 = np.zeros(sqw0.shape)

    sqw = np.vstack((sqw0, sqw1, sqw1))

    return sqw


def PrefDemoQscan(H, K, L, W, EXP, p):
    #Prefactor example for convolution tests
    
    [sample, rsample] = EXP.get_lattice()

    q_2 = npy.tools._modvec([H, K, L], rsample) ** 2

    intensity = p[5]
    bgr       = p[6]
    #temp = p[7]
    #Mn2+
    AA =   p[8]
    aa =   p[9]
    BB =   p[10]
    bb =   p[11]
    CC =   p[12]
    cc =   p[13]
    DD =   p[14]

    # Now, use the Jane Brown approximate expression for Co2+
    sd = q_2 / (16 * np.pi ** 2)
    ff = AA*np.exp(-aa*sd) + BB*np.exp(-bb*sd) + CC*np.exp(-cc*sd) + DD

    alphax = angle2(1, 0, 0, H, K, L, sample)
    alphay = angle2(0, 1, 0, H, K, L, sample)
    alphaz = angle2(0, 0, 1, H, K, L, sample)

    polx = np.sin(alphax) ** 2
    poly = np.sin(alphay) ** 2
    polz = np.sin(alphaz) ** 2

    prefactor = np.zeros((3, len(H)))
    prefactor[0, :] = ff ** 2.0 * polx * intensity
    prefactor[1, :] = 0 
    prefactor[2, :] = 0 

    bgr = np.ones(H.shape) * bgr

    return [prefactor, bgr]

        
       
def ConvResidualDemo(parameters, hklw, exp):

    p1      =  float(parameters['p1'])
    p2      =  float(parameters['p2'])
    w1      =  float(parameters['w1'])
    w2      =  float(parameters['w2'])
    ratio   =  float(parameters['ratio'])
    height  =  float(parameters['height'])
    bg      =  float(parameters['bg'])
    temp    =  float(parameters['temp'])

    parlist =  [p1, p2,  w1, w2, ratio, height, bg, temp]

    conv    =  exp.ResConv(SqwDemo, PrefDemo, nargout=2, hkle=hklw[0:4,:], METHOD='fix', ACCURACY=[5,5], p=parlist)
    residual = conv-hklw[5,:]

    return residual

def save_fit_params(scanno, hklw, finalpar, outfile="fittedparams.txt"):
    
    with open(outfile, 'a+') as f:
        f.write("Fitting to scan #{}_at_({}, {}, {})\n".format(scanno, hklw['h'][0], hklw['k'][0], hklw['l'][0]))
        f.write("    e1      e1err      e2      e2err      ratio      rtoerr      w1      w1err      w2      w2err      intensity      interr      bg      bgerr\n")
        f.write("{:6.4f}    {:6.4f}    {:6.4f}    {:6.4f}    {:6.4f}    {:6.4f}    {:6.4f}    {:6.4f}    {:6.4f}    {:6.4f}    {:6.4f}    {:6.4f}    {:6.4f}    {:6.4f}\n".format(finalpar['e1'][0],finalpar['e1err'][0],finalpar['e2'][0],
             finalpar['e2err'][0],finalpar['ratio'][0], finalpar['rtoerr'][0], finalpar['w1'][0], finalpar['w1err'][0],finalpar['w2'][0],finalpar['w2err'][0],finalpar['int'][0],finalpar['interr'][0],finalpar['bg'][0],finalpar['bgerr'][0]))
        f.write("\n")

    print("The parameters are written in the file {}!".format(outfile))
    return

def SelFormFactor(magion=''):
    formfactList = [
        {"magion": "Sc0", "AA": 0.251200, "aa": 90.029600, "BB": 0.329000, "bb": 39.402100, "CC": 0.423500, "cc": 14.322200, "DD": -0.004300 },
        {"magion": "Sc1", "AA": 0.488900, "aa": 51.160300, "BB": 0.520300, "bb": 14.076400, "CC": -0.028600, "cc": 0.179200, "DD": 0.018500 },
        {"magion": "Sc2", "AA": 0.504800, "aa": 31.403500, "BB": 0.518600, "bb": 10.989700, "CC": -0.024100, "cc": 1.183100, "DD": 0.000000 },
        {"magion": "Ti0", "AA": 0.465700, "aa": 33.589800, "BB": 0.549000, "bb": 9.879100, "CC": -0.029100, "cc": 0.323200, "DD": 0.012300 },
        {"magion": "Ti1", "AA": 0.509300, "aa": 36.703300, "BB": 0.503200, "bb": 10.371300, "CC": -0.026300, "cc": 0.310600, "DD": 0.011600 },
        {"magion": "Ti2", "AA": 0.509100, "aa": 24.976300, "BB": 0.516200, "bb": 8.756900, "CC": -0.028100, "cc": 0.916000, "DD": 0.001500 },
        {"magion": "Ti3", "AA": 0.357100, "aa": 22.841300, "BB": 0.668800, "bb": 8.930600, "CC": -0.035400, "cc": 0.483300, "DD": 0.009900 },
        {"magion": "V0", "AA": 0.408600, "aa": 28.810900, "BB": 0.607700, "bb": 8.543700, "CC": -0.029500, "cc": 0.276800, "DD": 0.012300 },
        {"magion": "V1", "AA": 0.444400, "aa": 32.647900, "BB": 0.568300, "bb": 9.097100, "CC": -0.228500, "cc": 0.021800, "DD": 0.215000 },
        {"magion": "V2", "AA": 0.408500, "aa": 23.852600, "BB": 0.609100, "bb": 8.245600, "CC": -0.167600, "cc": 0.041500, "DD": 0.149600 },
        {"magion": "V3", "AA": 0.359800, "aa": 19.336400, "BB": 0.663200, "bb": 7.617200, "CC": -0.306400, "cc": 0.029600, "DD": 0.283500 },
        {"magion": "V4", "AA": 0.310600, "aa": 16.816000, "BB": 0.719800, "bb": 7.048700, "CC": -0.052100, "cc": 0.302000, "DD": 0.022100 },
        {"magion": "Cr0", "AA": 0.113500, "aa": 45.199000, "BB": 0.348100, "bb": 19.493100, "CC": 0.547700, "cc": 7.354200, "DD": -0.009200 },
        {"magion": "Cr1", "AA": -0.097700, "aa": 0.047000, "BB": 0.454400, "bb": 26.005400, "CC": 0.557900, "cc": 7.489200, "DD": 0.083100 },
        {"magion": "Cr2", "AA": 1.202400, "aa": -0.005500, "BB": 0.415800, "bb": 20.547500, "CC": 0.603200, "cc": 6.956000, "DD": -1.221800 },
        {"magion": "Cr3", "AA": -0.309400, "aa": 0.027400, "BB": 0.368000, "bb": 17.035500, "CC": 0.655900, "cc": 6.523600, "DD": 0.285600 },
        {"magion": "Cr4", "AA": -0.232000, "aa": 0.043300, "BB": 0.310100, "bb": 14.951800, "CC": 0.718200, "cc": 6.172600, "DD": 0.204200 },
        {"magion": "Mn0", "AA": 0.243800, "aa": 24.962900, "BB": 0.147200, "bb": 15.672800, "CC": 0.618900, "cc": 6.540300, "DD": -0.010500 },
        {"magion": "Mn1", "AA": -0.013800, "aa": 0.421300, "BB": 0.423100, "bb": 24.668000, "CC": 0.590500, "cc": 6.654500, "DD": -0.001000 },
        {"magion": "Mn2", "AA": 0.422000, "aa": 17.684000, "BB": 0.594800, "bb": 6.005000, "CC": 0.004300, "cc": -0.609000, "DD": -0.021900 },
        {"magion": "Mn3", "AA": 0.419800, "aa": 14.282900, "BB": 0.605400, "bb": 5.468900, "CC": 0.924100, "cc": -0.008800, "DD": -0.949800 },
        {"magion": "Mn4", "AA": 0.376000, "aa": 12.566100, "BB": 0.660200, "bb": 5.132900, "CC": -0.037200, "cc": 0.563000, "DD": 0.001100 },
        {"magion": "Mn5", "AA": 0.292400, "aa": 11.665500, "BB": 0.740500, "bb": 5.074100, "CC": -1.788300, "cc": 0.005900, "DD": 1.755700 },
        {"magion": "Fe0", "AA": 0.070600, "aa": 35.008500, "BB": 0.358900, "bb": 15.358300, "CC": 0.581900, "cc": 5.560600, "DD": -0.011400 },
        {"magion": "Fe1", "AA": 0.125100, "aa": 34.963300, "BB": 0.362900, "bb": 15.514400, "CC": 0.522300, "cc": 5.591400, "DD": -0.010500 },
        {"magion": "Fe2", "AA": 0.026300, "aa": 34.959700, "BB": 0.366800, "bb": 15.943500, "CC": 0.618800, "cc": 5.593500, "DD": -0.011900 },
        {"magion": "Fe3", "AA": 0.397200, "aa": 13.244200, "BB": 0.629500, "bb": 4.903400, "CC": -0.031400, "cc": 0.349600, "DD": 0.004400 },
        {"magion": "Fe4", "AA": 0.378200, "aa": 11.380000, "BB": 0.655600, "bb": 4.592000, "CC": -0.034600, "cc": 0.483300, "DD": 0.000500 },
        {"magion": "Co0", "AA": 0.413900, "aa": 16.161600, "BB": 0.601300, "bb": 4.780500, "CC": -0.151800, "cc": 0.021000, "DD": 0.134500 },
        {"magion": "Co1", "AA": 0.099000, "aa": 33.125200, "BB": 0.364500, "bb": 15.176800, "CC": 0.547000, "cc": 5.008100, "DD": -0.010900 },
        {"magion": "Co2", "AA": 0.433200, "aa": 14.355300, "BB": 0.585700, "bb": 4.607700, "CC": -0.038200, "cc": 0.133800, "DD": 0.017900 },
        {"magion": "Co3", "AA": 0.390200, "aa": 12.507800, "BB": 0.632400, "bb": 4.457400, "CC": -0.150000, "cc": 0.034300, "DD": 0.127200 },
        {"magion": "Co4", "AA": 0.351500, "aa": 10.778500, "BB": 0.677800, "bb": 4.234300, "CC": -0.038900, "cc": 0.240900, "DD": 0.009800 },
        {"magion": "Ni0", "AA": -0.017200, "aa": 35.739200, "BB": 0.317400, "bb": 14.268900, "CC": 0.713600, "cc": 4.566100, "DD": -0.014300 },
        {"magion": "Ni1", "AA": 0.070500, "aa": 35.856100, "BB": 0.398400, "bb": 13.804200, "CC": 0.542700, "cc": 4.396500, "DD": -0.011800 },
        {"magion": "Ni2", "AA": 0.016300, "aa": 35.882600, "BB": 0.391600, "bb": 13.223300, "CC": 0.605200, "cc": 4.338800, "DD": -0.013300 },
        {"magion": "Ni3", "AA": 0.001200, "aa": 34.999800, "BB": 0.346800, "bb": 11.987400, "CC": 0.666700, "cc": 4.251800, "DD": -0.014800 },
        {"magion": "Ni4", "AA": -0.009000, "aa": 35.861400, "BB": 0.277600, "bb": 11.790400, "CC": 0.747400, "cc": 4.201100, "DD": -0.016300 },
        {"magion": "Cu0", "AA": 0.090900, "aa": 34.983800, "BB": 0.408800, "bb": 11.443200, "CC": 0.512800, "cc": 3.824800, "DD": -0.012400 },
        {"magion": "Cu1", "AA": 0.074900, "aa": 34.965600, "BB": 0.414700, "bb": 11.764200, "CC": 0.523800, "cc": 3.849700, "DD": -0.012700 },
        {"magion": "Cu2", "AA": 0.023200, "aa": 34.968600, "BB": 0.402300, "bb": 11.564000, "CC": 0.588200, "cc": 3.842800, "DD": -0.013700 },
        {"magion": "Cu3", "AA": 0.003100, "aa": 34.907400, "BB": 0.358200, "bb": 10.913800, "CC": 0.653100, "cc": 3.827900, "DD": -0.014700 },
        {"magion": "Cu4", "AA": -0.013200, "aa": 30.681700, "BB": 0.280100, "bb": 11.162600, "CC": 0.749000, "cc": 3.817200, "DD": -0.016500 },
        {"magion": "Y0", "AA": 0.591500, "aa": 67.608100, "BB": 1.512300, "bb": 17.900400, "CC": -1.113000, "cc": 14.135900, "DD": 0.008000 },
        {"magion": "Zr0", "AA": 0.410600, "aa": 59.996100, "BB": 1.054300, "bb": 18.647600, "CC": -0.475100, "cc": 10.540000, "DD": 0.010600 },
        {"magion": "Zr1", "AA": 0.453200, "aa": 59.594800, "BB": 0.783400, "bb": 21.435700, "CC": -0.245100, "cc": 9.036000, "DD": 0.009800 },
        {"magion": "Nb0", "AA": 0.394600, "aa": 49.229700, "BB": 1.319700, "bb": 14.821600, "CC": -0.726900, "cc": 9.615600, "DD": 0.012900 },
        {"magion": "Nb1", "AA": 0.457200, "aa": 49.918200, "BB": 1.027400, "bb": 15.725600, "CC": -0.496200, "cc": 9.157300, "DD": 0.011800 },
        {"magion": "Mo0", "AA": 0.180600, "aa": 49.056800, "BB": 1.230600, "bb": 14.785900, "CC": -0.426800, "cc": 6.986600, "DD": 0.017100 },
        {"magion": "Mo1", "AA": 0.350000, "aa": 48.035400, "BB": 1.030500, "bb": 15.060400, "CC": -0.392900, "cc": 7.479000, "DD": 0.013900 },
        {"magion": "Tc0", "AA": 0.129800, "aa": 49.661100, "BB": 1.165600, "bb": 14.130700, "CC": -0.313400, "cc": 5.512900, "DD": 0.019500 },
        {"magion": "Tc1", "AA": 0.267400, "aa": 48.956600, "BB": 0.956900, "bb": 15.141300, "CC": -0.238700, "cc": 5.457800, "DD": 0.016000 },
        {"magion": "Ru0", "AA": 0.106900, "aa": 49.423800, "BB": 1.191200, "bb": 12.741700, "CC": -0.317600, "cc": 4.912500, "DD": 0.021300 },
        {"magion": "Ru1", "AA": 0.441000, "aa": 33.308600, "BB": 1.477500, "bb": 9.553100, "CC": -0.936100, "cc": 6.722000, "DD": 0.017600 },
        {"magion": "Rh0", "AA": 0.097600, "aa": 49.882500, "BB": 1.160100, "bb": 11.830700, "CC": -0.278900, "cc": 4.126600, "DD": 0.023400 },
        {"magion": "Rh1", "AA": 0.334200, "aa": 29.756400, "BB": 1.220900, "bb": 9.438400, "CC": -0.575500, "cc": 5.332000, "DD": 0.021000 },
        {"magion": "Pd0", "AA": 0.200300, "aa": 29.363300, "BB": 1.144600, "bb": 9.599300, "CC": -0.368900, "cc": 4.042300, "DD": 0.025100 },
        {"magion": "Pd1", "AA": 0.503300, "aa": 24.503700, "BB": 1.998200, "bb": 6.908200, "CC": -1.524000, "cc": 5.513300, "DD": 0.021300 },
        {"magion": "Ce2", "AA": 0.295300, "aa": 17.684600, "BB": 0.292300, "bb": 6.732900, "CC": 0.431300, "cc": 5.382700, "DD": -0.019400 },
        {"magion": "Nd2", "AA": 0.164500, "aa": 25.045300, "BB": 0.252200, "bb": 11.978200, "CC": 0.601200, "cc": 4.946100, "DD": -0.018000 },
        {"magion": "Nd3", "AA": 0.054000, "aa": 25.029300, "BB": 0.310100, "bb": 12.102000, "CC": 0.657500, "cc": 4.722300, "DD": -0.021600 },
        {"magion": "Sm2", "AA": 0.090900, "aa": 25.203200, "BB": 0.303700, "bb": 11.856200, "CC": 0.625000, "cc": 4.236600, "DD": -0.020000 },
        {"magion": "Sm3", "AA": 0.028800, "aa": 25.206800, "BB": 0.297300, "bb": 11.831100, "CC": 0.695400, "cc": 4.211700, "DD": -0.021300 },
        {"magion": "Eu2", "AA": 0.075500, "aa": 25.296000, "BB": 0.300100, "bb": 11.599300, "CC": 0.643800, "cc": 4.025200, "DD": -0.019600 },
        {"magion": "Eu3", "AA": 0.020400, "aa": 25.307800, "BB": 0.301000, "bb": 11.474400, "CC": 0.700500, "cc": 3.942000, "DD": -0.022000 },
        {"magion": "Gd2", "AA": 0.063600, "aa": 25.382300, "BB": 0.303300, "bb": 11.212500, "CC": 0.652800, "cc": 3.787700, "DD": -0.019900 },
        {"magion": "Gd3", "AA": 0.018600, "aa": 25.386700, "BB": 0.289500, "bb": 11.142100, "CC": 0.713500, "cc": 3.752000, "DD": -0.021700 },
        {"magion": "Tb2", "AA": 0.054700, "aa": 25.508600, "BB": 0.317100, "bb": 10.591100, "CC": 0.649000, "cc": 3.517100, "DD": -0.021200 },
        {"magion": "Tb3", "AA": 0.017700, "aa": 25.509500, "BB": 0.292100, "bb": 10.576900, "CC": 0.713300, "cc": 3.512200, "DD": -0.023100 },
        {"magion": "Dy2", "AA": 0.130800, "aa": 18.315500, "BB": 0.311800, "bb": 7.664500, "CC": 0.579500, "cc": 3.146900, "DD": -0.022600 },
        {"magion": "Dy3", "AA": 0.115700, "aa": 15.073200, "BB": 0.327000, "bb": 6.799100, "CC": 0.582100, "cc": 3.020200, "DD": -0.024900 },
        {"magion": "Ho2", "AA": 0.099500, "aa": 18.176100, "BB": 0.330500, "bb": 7.855600, "CC": 0.592100, "cc": 2.979900, "DD": -0.023000 },
        {"magion": "Ho3", "AA": 0.056600, "aa": 18.317600, "BB": 0.336500, "bb": 7.688000, "CC": 0.631700, "cc": 2.942700, "DD": -0.024800 },
        {"magion": "Er2", "AA": 0.112200, "aa": 18.122300, "BB": 0.346200, "bb": 6.910600, "CC": 0.564900, "cc": 2.761400, "DD": -0.023500 },
        {"magion": "Er3", "AA": 0.058600, "aa": 17.980200, "BB": 0.354000, "bb": 7.096400, "CC": 0.612600, "cc": 2.748200, "DD": -0.025100 },
        {"magion": "Tm2", "AA": 0.098300, "aa": 18.323600, "BB": 0.338000, "bb": 6.917800, "CC": 0.587500, "cc": 2.662200, "DD": -0.024100 },
        {"magion": "Tm3", "AA": 0.058100, "aa": 15.092200, "BB": 0.278700, "bb": 7.801500, "CC": 0.685400, "cc": 2.793100, "DD": -0.022400 },
        {"magion": "Yb2", "AA": 0.085500, "aa": 18.512300, "BB": 0.294300, "bb": 7.373400, "CC": 0.641200, "cc": 2.677700, "DD": -0.021300 },
        {"magion": "Yb3", "AA": 0.041600, "aa": 16.094900, "BB": 0.284900, "bb": 7.834100, "CC": 0.696100, "cc": 2.672500, "DD": -0.022900 },
        {"magion": "Pr3", "AA": 0.050400, "aa": 24.998900, "BB": 0.257200, "bb": 12.037700, "CC": 0.714200, "cc": 5.003900, "DD": -0.021900 },
        {"magion": "U3", "AA": 0.505800, "aa": 23.288200, "BB": 1.346400, "bb": 7.002800, "CC": -0.872400, "cc": 4.868300, "DD": 0.019200 },
        {"magion": "U4", "AA": 0.329100, "aa": 23.547500, "BB": 1.083600, "bb": 8.454000, "CC": -0.434000, "cc": 4.119600, "DD": 0.021400 },
        {"magion": "U5", "AA": 0.365000, "aa": 19.803800, "BB": 3.219900, "bb": 6.281800, "CC": -2.607700, "cc": 5.301000, "DD": 0.023300 },
        {"magion": "Np3", "AA": 0.515700, "aa": 20.865400, "BB": 2.278400, "bb": 5.893000, "CC": -1.816300, "cc": 4.845700, "DD": 0.021100 },
        {"magion": "Np4", "AA": 0.420600, "aa": 19.804600, "BB": 2.800400, "bb": 5.978300, "CC": -2.243600, "cc": 4.984800, "DD": 0.022800 },
        {"magion": "Np5", "AA": 0.369200, "aa": 18.190000, "BB": 3.151000, "bb": 5.850000, "CC": -2.544600, "cc": 4.916400, "DD": 0.024800 },
        {"magion": "Np6", "AA": 0.292900, "aa": 17.561100, "BB": 3.486600, "bb": 5.784700, "CC": -2.806600, "cc": 4.870700, "DD": 0.026700 },
        {"magion": "Pu3", "AA": 0.384000, "aa": 16.679300, "BB": 3.104900, "bb": 5.421000, "CC": -2.514800, "cc": 4.551200, "DD": 0.026300 },
        {"magion": "Pu4", "AA": 0.493400, "aa": 16.835500, "BB": 1.639400, "bb": 5.638400, "CC": -1.158100, "cc": 4.139900, "DD": 0.024800 },
        {"magion": "Pu5", "AA": 0.388800, "aa": 16.559200, "BB": 2.036200, "bb": 5.656700, "CC": -1.451500, "cc": 4.255200, "DD": 0.026700 },
        {"magion": "Pu6", "AA": 0.317200, "aa": 16.050700, "BB": 3.465400, "bb": 5.350700, "CC": -2.810200, "cc": 4.513300, "DD": 0.028100 },
        {"magion": "Am2", "AA": 0.474300, "aa": 21.776100, "BB": 1.580000, "bb": 5.690200, "CC": -1.077900, "cc": 4.145100, "DD": 0.021800 },
        {"magion": "Am3", "AA": 0.423900, "aa": 19.573900, "BB": 1.457300, "bb": 5.872200, "CC": -0.905200, "cc": 3.968200, "DD": 0.023800 },
        {"magion": "Am4", "AA": 0.373700, "aa": 17.862500, "BB": 1.352100, "bb": 6.042600, "CC": -0.751400, "cc": 3.719900, "DD": 0.025800 },
        {"magion": "Am5", "AA": 0.295600, "aa": 17.372500, "BB": 1.452500, "bb": 6.073400, "CC": -0.775500, "cc": 3.661900, "DD": 0.027700 },
        {"magion": "Am6", "AA": 0.230200, "aa": 16.953300, "BB": 1.486400, "bb": 6.115900, "CC": -0.745700, "cc": 3.542600, "DD": 0.029400 },
        {"magion": "Am7", "AA": 0.360100, "aa": 12.729900, "BB": 1.964000, "bb": 5.120300, "CC": -1.356000, "cc": 3.714200, "DD": 0.031600 }
    ]
    
    for item in formfactList:
        if magion.upper() == item["magion"].upper():
            #print(f"The form factor of {magion} is found.")
            return item
    print(f"The mangetic ion {magion} is not found.")
    return None


def curie_weiss_law(T, par_C, par_Tc):
    inv_kai= T/par_C - par_Tc/par_C
    return inv_kai

    
def residual_curie_weiss(pars, x, data=None):
    """Model """
    vals = pars.valuesdict()
    par_C   = vals['C']
    par_Tc  = vals['Tc']
    
    model = x/par_C - par_Tc/par_C
    if data is None:
        return model
    return model - data



def create_html_plots(dataframes_config, output_file="output_dashboard.html", cols=4):
    """
    Create an HTML file with plotly visualizations for multiple dataframes
    
    Parameters:
    dataframes_config (list): List of dictionaries with format:
                             {'filename': filename, 
                              'df': dataframe, 
                              'x_col': x_axis_column,
                              'y_cols': [list_of_y_columns],
                              'plot_type': 'line'/'bar'/'scatter'/etc.}
    output_file (str): Path to save the HTML output
    cols (int): Number of columns in the layout
    """
    # Sort dataframes by modification time (most recent first)
    for config in dataframes_config:
        try:
            config['mod_time'] = os.path.getmtime(config['filename'])
        except:
            # If filename isn't a real file path, use current time
            config['mod_time'] = datetime.now().timestamp()
    
    # Sort by modification time (most recent first)
    dataframes_config.sort(key=lambda x: x['mod_time'], reverse=True)
    
    # Create HTML file
    with open(output_file, 'w') as f:
        # HTML header
        f.write("""
        <!DOCTYPE html>
        <html>
        <head>
            <title>Data Visualization Dashboard</title>
            <script src="https://cdn.plot.ly/plotly-latest.min.js"></script>
            <style>
                body {
                    font-family: Arial, sans-serif;
                    margin: 20px;
                }
                .grid-container {
                    display: grid;
                    grid-template-columns: repeat(""" + str(cols) + """, 1fr);
                    gap: 20px;
                    padding: 20px;
                }
                .plot-container {
                    border: 1px solid #ddd;
                    border-radius: 5px;
                    padding: 15px;
                    box-shadow: 0 2px 5px rgba(0,0,0,0.1);
                }
                h1 {
                    text-align: center;
                }
                h3 {
                    margin-top: 0;
                    color: #333;
                }
                @media (max-width: 1200px) {
                    .grid-container {
                        grid-template-columns: repeat(3, 1fr);
                    }
                }
                @media (max-width: 900px) {
                    .grid-container {
                        grid-template-columns: repeat(2, 1fr);
                    }
                }
                @media (max-width: 600px) {
                    .grid-container {
                        grid-template-columns: 1fr;
                    }
                }
            </style>
        </head>
        <body>
            <h1>Data Visualization Dashboard</h1>
            <div class="grid-container">
        """)
        
        # Add each plot
        for config in dataframes_config:
            plot_div = generate_plotly(
                config['df'], 
                os.path.basename(config['filename']),
                config.get('x_col'),
                config.get('y_cols', []),
                config.get('plot_type', 'line')
            )
            
            # Write plot container
            f.write(f"""
            <div class="plot-container">
                <h3>{os.path.basename(config['filename'])}</h3>
                {plot_div}
            </div>
            """)
        
        # HTML footer
        f.write("""
            </div>
        </body>
        </html>
        """)
    
    print(f"Dashboard created successfully at {output_file}")

def generate_plotly(df, title, x_col=None, y_cols=None, plot_type='line'):
    """
    Generate a plotly visualization for a dataframe
    
    Parameters:
    df (DataFrame): The pandas DataFrame to plot
    title (str): Title for the plot
    x_col (str): Column name to use for x-axis
    y_cols (list): List of column names to use for y-axis
    plot_type (str): Type of plot ('line', 'bar', 'scatter', etc.)
    """
    fig = go.Figure()
    
    # Set default x and y columns if not specified
    if x_col is None:
        x_col = df.columns[0] if len(df.columns) > 0 else None
    
    if y_cols is None or len(y_cols) == 0:
        if len(df.columns) > 1:
            y_cols = df.columns[1:]
        else:
            y_cols = [df.columns[0]] if len(df.columns) > 0 else []
    
    # Create the plot based on specified type
    for y_col in y_cols:
        if plot_type == 'line':
            fig.add_trace(go.Scatter(x=df[x_col], y=df[y_col], mode='lines+markers', name=y_col))
        elif plot_type == 'bar':
            fig.add_trace(go.Bar(x=df[x_col], y=df[y_col], name=y_col))
        elif plot_type == 'scatter':
            fig.add_trace(go.Scatter(x=df[x_col], y=df[y_col], mode='markers', name=y_col))
        elif plot_type == 'area':
            fig.add_trace(go.Scatter(x=df[x_col], y=df[y_col], fill='tozeroy', name=y_col))
        elif plot_type == 'box':
            fig.add_trace(go.Box(y=df[y_col], name=y_col))
        elif plot_type == 'histogram':
            fig.add_trace(go.Histogram(x=df[y_col], name=y_col))
        else:  # Default to line plot
            fig.add_trace(go.Scatter(x=df[x_col], y=df[y_col], mode='lines+markers', name=y_col))
    
    # Update layout
    fig.update_layout(
        title=title,
        xaxis_title=x_col,
        yaxis_title="Value",
        height=400,
        margin=dict(l=10, r=10, t=40, b=10),
        template="plotly_white",
        legend=dict(orientation="h", yanchor="bottom", y=1.02, xanchor="right", x=1)
    )
    
    # Convert plot to HTML div
    plot_div = pyo.plot(fig, output_type='div', include_plotlyjs=False)
    
    return plot_div




# Create a contour plot using matplotlib
def matplotlib_contour(data, xtitle='x', ytitle= 'y', ztitle= 'counts', zminmax=[0, 1000], output_file=None):
    # Extract data
    x = data[xtitle].values
    y = data[ytitle].values
    z = data[ztitle].values
    
    # Create a regular grid for contour plot
    xi = np.linspace(min(x), max(x), 100)
    yi = np.linspace(min(y), max(y), 100)
    xi, yi = np.meshgrid(xi, yi)
    
    # Interpolate z values on the regular grid
    #zi = griddata((x, y), z, (xi, yi), method='cubic')
    # Interpolate z values on the regular grid
    zi = griddata((x, y), z, (xi, yi), method='linear')  # Change to 'linear' to avoid extrapolation issues
    zi = np.clip(zi, zminmax[0], zminmax[1])  # Clip values to defined range
    zi = np.nan_to_num(zi, nan=zminmax[0])  # Replace NaN with min value
    
    # Create the contour plot
    fig, ax = plt.subplots(figsize=(10,8))
    contour = plt.contourf(xi, yi, zi, levels=15, vmin=zminmax[0], vmax=zminmax[1], cmap='viridis')
    plt.colorbar(contour, label='Counts')
    plt.scatter(x, y, s=1, color='black', alpha=0.5)  # Show original data points
    plt.title('Contour Map of Measurement Data')
    plt.xlabel('X Position')
    plt.ylabel('Y Position')
    plt.grid(True, linestyle='--', alpha=0.7)
    if output_file:
        plt.savefig('contour_plot_matplotlib.png', dpi=300, bbox_inches='tight')
    #plt.show()
    return fig, ax



def plotly_contour(data, xtitle='x', ytitle= 'y', ztitle= 'counts', zminmax=[0, 1000], output_file=None):
    # Extract data
    x = data[xtitle].values
    y = data[ytitle].values
    z = data[ztitle].values
    
    # Create a regular grid for contour plot
    xi = np.linspace(min(x), max(x), 100)
    yi = np.linspace(min(y), max(y), 100)
    xi, yi = np.meshgrid(xi, yi)
    
    # Interpolate z values on the regular grid
    zi = griddata((x, y), z, (xi, yi), method='cubic')
    
    # Create the contour plot
    fig = go.Figure(data=go.Contour(
        x=xi[0],
        y=yi[:,0],
        z=zi,
        colorscale='Viridis',
        contours=dict(
            showlabels=True,
            labelfont=dict(size=12, color='white')
        ),zmin=zminmax[0],zmax=zminmax[1], ncontours=20,
        colorbar=dict(title='Counts'),
        hoverinfo='x+y+z'
    ))
    
    # Add scatter points for original data
    fig.add_trace(go.Scatter(
        x=x,
        y=y,
        mode='markers',
        marker=dict(
            size=3,
            color='black',
            opacity=0.5
        ),
        name='Data Points',
        hoverinfo='x+y+text',
        hovertext=[f'Count: {count:.2f}' for count in z]
    ))
    
    # Update layout
    fig.update_layout(
        title='Contour Map of Measurement Data',
        xaxis_title='X Position',
        yaxis_title='Y Position',
        width=900,
        height=700
    )
    
    # Save to HTML file if specified
    if output_file:
        fig.write_html(output_file)
    

    return fig    


def set_standard_plot(ax, title, xlabel="X-axis", ylabel="Y-axis", bGrid=False):
    # Title
    ax.set_title(title, fontsize=14, fontweight='bold', pad=10)
    
    # Labels
    ax.set_xlabel(xlabel, fontsize=18, labelpad=10)
    ax.set_ylabel(ylabel, fontsize=18, labelpad=10)
    
    # Tick labels
    ax.tick_params(axis='both', which='major', labelsize=14, width=2, length=6, pad=5)
    ax.tick_params(axis='both', which='minor', width=1.5, length=4)
    
    # Tick spacing
    #ax.xaxis.set_major_locator(MultipleLocator(2))
    #ax.xaxis.set_minor_locator(MultipleLocator(0.5))
    #ax.yaxis.set_major_locator(MultipleLocator(0.5))
    #ax.yaxis.set_minor_locator(MultipleLocator(0.1))
    # Frame appearance
    for spine in ax.spines.values():
        spine.set_linewidth(2)

    # Grid
    if bGrid:
        ax.grid(True, linestyle='--', linewidth=0.8, alpha=0.7)
    
    # Legend
    #ax.legend(fontsize=12, frameon=True, fancybox=True, framealpha=0.8, 
    #          edgecolor='black', borderpad=1, handlelength=3)
    #ax.text(0.02, 0.98, '(a)', transform=ax.transAxes,
    #    fontsize=14,  va='top', ha='left') #fontweight='bold',


def customize_plot_axis(ax, title, xlabel="X-axis", ylabel="Y-axis"):
    # Title
    ax.set_title(title, fontsize=14, fontweight='bold', pad=10)
    
    # Labels
    ax.set_xlabel(xlabel, fontsize=18, labelpad=10)
    ax.set_ylabel(ylabel, fontsize=18, labelpad=10)
    
    # Tick labels
    ax.tick_params(axis='both', which='major', labelsize=14, width=2, length=6, pad=5)
    ax.tick_params(axis='both', which='minor', width=1.5, length=4)
    
    # Tick spacing
    #ax.xaxis.set_major_locator(MultipleLocator(2))
    #ax.xaxis.set_minor_locator(MultipleLocator(0.5))
    #ax.yaxis.set_major_locator(MultipleLocator(0.5))
    #ax.yaxis.set_minor_locator(MultipleLocator(0.1))
    
    # Frame appearance
    for spine in ax.spines.values():
        spine.set_linewidth(2)
    
    # Grid
    #ax.grid(True, linestyle='--', linewidth=0.8, alpha=0.7)
    
    # Legend
    ax.legend(fontsize=12, frameon=True, fancybox=True, framealpha=0.8, 
              edgecolor='black', borderpad=1, handlelength=3)
    ax.text(0.02, 0.98, '(a)', transform=ax.transAxes,
        fontsize=14,  va='top', ha='left') #fontweight='bold',


from pathlib import Path
import pandas as pd
import numpy as np
import re

import inspy as npy
from inspy import TripleAxisSpectr
from inspy.insfit import FitConv, UltraFastFitConv


from ..base import TasData
from .validator import SikaCommandValidator
from ..utils.toolfunc import (AnglesToQhkl, strisfloat, strisint,  
                gaussian_residual, lorentzian_residual, fit_peak, 
                angle2, SqwDemo, PrefDemo, SelFormFactor, PrefPhononDemo)



class Sika(TasData):
    """Sika class that extends TasData."""
    
    def __init__(self, expnum, title, sample, user):
        super().__init__("Sika", expnum, title, sample, user)
        #self.specific_value = specific_value
    

    def _sika_data_header_info(self, path):
        """
        Parse a neutron data file and extract header information.
        Args:
            path (str): Path to the data file
        Returns:
            dict: Dictionary containing all extracted parameters
        """
        # Dictionary to store all extracted parameters
        params = {}
        
        # Keep track of last header line and data start line
        last_header_line = ""
        last_header_line_num = 0
        data_start_line = 0
        line_count = 0
        
        # Special handling for arrays
        array_params = ['latticeconstants', 'ubmatrix', 'plane_normal']
        
        with open(path, 'r') as f:
            for line in f:
                line_count += 1
                
                if line.startswith('#'):
                    last_header_line = line
                    last_header_line_num = line_count
                    
                    # Skip the header line that contains column names
                    if line.strip().startswith('# Pt.') or '# Pt.' in line:
                        continue
                    
                    # Process parameter line
                    clean_line = line[1:].strip()
                    
                    # Skip empty lines or lines without '='
                    if not clean_line or '=' not in clean_line:
                        continue
                    
                    # Split by the first occurrence of '='
                    key, value = clean_line.split('=', 1)
                    key = key.strip()
                    value = value.strip()
                    
                    # Handle special array parameters
                    if key in array_params:
                        # Split by commas and convert to float
                        value = [float(item) for item in value.split(',')]
                    
                    # Store in dictionary
                    params[key] = value
                else:
                    # First non-comment line is the start of data
                    data_start_line = line_count
                    break
        
        # Add special information about the file structure
        params['_data_start_line'] = data_start_line
        params['_header_line']     = last_header_line
        params['_header_line_num'] = last_header_line_num

        params['expno']   = params.pop('experiment_number')
        params['scanno']  = params.pop('scan')
        params['lattice'] = params.pop('latticeconstants')
        
        if 'steps' in params:
            str_steps = params['steps']

            # Use regex to find all substrings before the colons
            matches = re.findall(r'([^:]+):', str_steps)
             
            if len(matches) == 1:
                params['scanax1']=matches[0]
                params['scanax2']=None
            elif len(matches) == 2:
                params['scanax1']=matches[0]
                params['scanax2']=matches[1]
            else:
                params['scanax1']=params['def_x']
                params['scanax2']=None
        else:
            params['scanax1']=params['def_x']
            params['scanax2']=None

        if params['scanax1'] == "":
            params['scanax1'] == "Pt."
            print("Scan axis is empty! Please check. it is temperarily set to Pt.")
        if params['scanax1'] in ['h', 'k', 'l']:
            params['scanax1']= "q"+ params['scanax1']
        elif params['scanax1'] in ['e']:
            params['scanax1']= "en"

            #print("no or too many scanning axis is found")


        return params


    def sika_data_to_pd(self, path='', scanno=None):
        """
        Load a single neutron data file into a pandas DataFrame with metadata in attrs.
        Args:
            path (str): Path to the data file
        Returns:
            pandas.DataFrame: DataFrame containing the data with metadata in attrs
        """
        ppath=Path(path)

        if not ppath.is_dir():
            print(ppath)
            print("Error: wrong path!")
            return

        if scanno is None:
            print("Error: No scan no. is given.")
            return

        basestr = "000000"
        basestrexp="0000"
        
        scanno = str(int(scanno)) 
        expno = str(int(self.expnum)) 
        filename = basestrexp[:-len(expno)]+expno  + "_" + basestr[:-len(scanno)] + scanno + '.dat'
        fullfilepath = ppath / Path(filename)

        # Parse header information
        params = self._sika_data_header_info(fullfilepath)
        #print(params)
        
        # Get file structure information
        data_start_line = params.pop('_data_start_line', 0)
        header_line = params.pop('_header_line', '')
        header_line_num = params.pop('_header_line_num', 0)
        
        # Process the header line to extract column names
        if header_line.startswith('#'):
            # Remove the leading # and split by whitespace
            header_line = header_line[1:].strip()
        # Create column names from the header line
        if header_line:
            # Clean and split the header line
            sika_column_names = [col.strip() for col in header_line.split() if col.strip()]
        #print(sika_column_names)
        sika_motorname_list = ['h', 'k', 'l','e', 'count', 'beamMon','tc1a','tc2a','magneta']
        stdtas_motorname_list = ['qh', 'qk', 'ql','en', 'detector', 'monitor', 'tempVTI', 'tempSAMP','magField']

        # Create a mapping from old to new
        replace_map = dict(zip(sika_motorname_list, stdtas_motorname_list))

        # change into standard column names
        std_column_names = [replace_map.get(item, item) for item in sika_column_names]

        # Try to read the data part using pandas
        try:
            # No header line found, use pandas to read with default column names
            df = pd.read_csv(fullfilepath, sep=r'\s+', header=None, comment='#')
            df.columns = std_column_names

        except Exception as e:
            print(f"Warning: Error parsing data section. Trying alternative method. Error: {e}")
            # Create an empty DataFrame as last resort
            df = pd.DataFrame()
    
        # Store all metadata in the DataFrame's attrs
        df.attrs = params
        
        return df  


    def sika_scanlist_to_dflist(self, path="", scanlist=None):

        dflist=[]

        if all(isinstance(x, (int, np.integer)) for x in scanlist):
            for scanno in scanlist:
                df  = self.sika_data_to_pd(path, scanno)
                dflist.append(df)

            #print(f"Totally, {len(dflist)} scans in the list are converted into df.")
            return dflist   
        else:
            print("Some element in the list is not integer. sublist in the list is not allowed.")
            return

    def sika_simpleplot(self, path="", scanlist=None, fit=False, initial=None):
        #quick plot without normalization and combination
        dflist = self.sika_batch_reduction(path, scanlist)
        #print(dflist)
        parlist, fitlist, fig, ax=self.tas_simpleplot(dflist, fit, initial)

        return parlist, fitlist, fig, ax


    def sika_combplot(self, path='', scanlist=None, fit=False, norm_mon_count = -999999, overplot=False, offset=1000, initial=None ):
        #scanlist can be [123, 125, [130, 132],115]
        dflist   = list()
        for scanno in scanlist:
            if isinstance(scanno, list):
                subscanlist = scanno
                subdflist   = self.sika_scanlist_to_dflist(path, subscanlist)
                comb_df  =  super().tas_datacombine(subdflist)
                dflist.append(comb_df)
            if isinstance(scanno, (int, np.integer)):
                single_df = self.sika_data_to_pd(path, scanno)
                dflist.append(single_df)
        
        params_df, fitted_df, fig, ax = super().tas_combplot(dflist, fit, norm_mon_count, overplot, offset, initial)

        return params_df, fitted_df, fig, ax

    def sika_combspline(self, path='', scanlist=None, fit=False, norm_mon_count = -999999, overplot=False, offset=1000, initial=None ):
        #scanlist can be [123, 125, [130, 132],115]
        dflist=[]
        for scanno in scanlist:
            if isinstance(scanno, list):
                subscanlist = scanno
                subdflist=self.sika_scanlist_to_dflist(path, subscanlist)
                comb_df=super().tas_datacombine(subdflist)
                #print("datacombined")
                dflist.append(comb_df)
                #print(comb_df)
            if isinstance(scanno, (int, np.integer)):
                single_df = self.sika_data_to_pd(path, scanno)
                dflist.append(single_df)
        #print(dflist)
        params_df, fitted_df, fig, ax=super().tas_combspline(dflist, fit, norm_mon_count, overplot, offset, initial)

        return params_df, fitted_df, fig, ax


    def sika_batch_reduction(self, path='', scanlist=None, motorlist=['qh', 'qk', 'ql', 'en', 'ei','ef', 'm1', 'm2','s1','s2', 'a1', 'a2', 'detector', 'monitor','tempVTI', 'tempSAMP']):
        #scanlist can be [123, 125, [130, 132],115]
        dflist     = list()
        col_names = motorlist #['h', 'k', 'l', 'e', 'q',  'm1', 'm2','s1','s2', 'a1', 'a2', 'sgl', 'sgu', 'stu', 'stl', 'sd','dd','i3h','i3v','i4h','i4v','Pt.','i1l','i1r', 'mfh', 'mfv', 'c4s','c4r', 'ei', 'ef', 'count', 'beamMon']
 
        for scanno in scanlist:
            #print(f"the problem here:{scanno}")
            #print(isinstance(scanno, int))  # this will be false, int is not np.integer the same
            if isinstance(scanno, (int, np.integer)):
                df = self.sika_data_to_pd(path, scanno)
            
            elif isinstance(scanno, list):
                subscanlist = scanno
                temp_dflist=self.sika_scanlist_to_dflist(path, subscanlist)
                df=super().tas_datacombine(temp_dflist)
            else:
                print("the elemenet in the list is not int or list.")
                return
            col_notexist=[]
            
            #print(df.columns)
            for col in col_names:
                if col not in df.columns:
                    col_notexist.append(col)
            # remove those col_names that is not in the data file in case. 
            col_names= [x for x in col_names if x not in col_notexist] 
            if col_notexist:
                print(f"These motors are not in the data:{col_notexist}")
            dflist.append(df[col_names])
                
        return dflist  # return reduced columns

    def sika_reduction_by_row(self, path='', scanlist=None, motorlist=['qh', 'qk', 'ql', 'en', 'ei','ef', 'm1', 'm2','s1','s2', 'a1', 'a2', 'detector', 'monitor','tempVTI', 'tempSAMP'], sortby="tempSAMP"):
        dflist = self.sika_batch_reduction(path, scanlist, motorlist)
        df_extend = pd.DataFrame(columns=motorlist)

        for df in dflist:
            if df is not None and not df.empty and not df.isna().all().all():
                df_extend=pd.concat([df_extend, df], ignore_index=True)
        if sortby in motorlist:
            df_extend = df_extend.sort_values(by=sortby, ascending = True).reset_index(drop=True)
        return df_extend

    def sika_reduction_anycolumn(self, path='', scanlist=None, motorlist=['qh', 'qk', 'ql', 'en', 'ei','ef', 'm1', 'm2','s1','s2', 'a1', 'a2', 'detector', 'monitor','tempVTI', 'tempSAMP'], sortby="tempSAMP"):
        dflist = self.sika_batch_reduction(path, scanlist, motorlist)
        df_extend = pd.DataFrame(columns=motorlist)

        for df in dflist:
            if df is not None and not df.empty and not df.isna().all().all():
                df_extend=pd.concat([df_extend, df], ignore_index=True)
        if sortby in motorlist:
            df_extend = df_extend.sort_values(by=sortby, ascending = True).reset_index(drop=True)
        return df_extend  

    def sika_random_contour(self, path='', scanlist=None, motorlist=['qh', 'qk', 'ql', 'en', 'detector', 'monitor'], x_col="qh", y_col="en", xtitle='qh [rlu]', ytitle='en [meV]',title='Contour Map of Measurement Data', vminmax=[0, 1000], output_file=None):
        df_total = self.sika_reduction_by_row(path, scanlist, motorlist=motorlist)
        xi, yi, zi, ax = super().tas_random_contour(df_total, x_col=x_col, y_col=y_col, xtitle=xtitle, ytitle=ytitle, title=title, vminmax=vminmax, output_file=output_file)
        return xi, yi, zi, ax

    def sika_tidy_contour(self, path='',  scanlist=None, motorlist=['qh', 'qk', 'ql', 'en', 'detector', 'monitor'], x_col='qh', y_col='en', xlabel='Q [rlu]',ylabel='E [meV]', vminmax=None, ax=None):
       
        dflist = self.sika_batch_reduction(path, scanlist, motorlist=motorlist)
        qq, ee, intt, ax = super().tas_tidy_contour(dflist,  x_col=x_col, y_col=y_col, xlabel=xlabel, ylabel=ylabel, vminmax=vminmax, ax=ax)
        return qq, ee, intt, ax


    def sika_export_hklw(self, path='', scanlist = None, hklw_file=""):
        #this is advanced export function, which can combine data and then export
        #scanlist could be [ 2, [3, 4], 5, 6] or [ 12, 23, 34, 45]
        dflist = self.sika_batch_reduction(path, scanlist, motorlist=['qh', 'qk', 'ql', 'en', 'detector'])
        
        super().tas_export_hklw(dflist, hklw_file)

        return dflist


    def export_scanlist(self, path='', datafromto=None, outputfile=''):
        # export the scan list from sika
        # path is the Datafiles folder
        # datafromto is a list with two elements, integers
        # outputfile is the filename for the list
        
        ppath=Path(path)

        if not ppath.is_dir():
            print("Error: wrong path!")
            return

        if datafromto is None:
            print("Error: No data from and to number is given")
            return
        elif datafromto[0] <= 0 or not isinstance(datafromto[0], int) or datafromto[1] <= 0 or not isinstance(datafromto[1], int):
            print("Error: Please specify positive integer for datafrom and to numbers!")
            return
        else:
            pass

        if outputfile:
            outputfile = ppath / Path(outputfile)
        else:
            outputfile = ppath / Path( "Sika_exp" + self.expnum  + "_scanfrom" + str(datafromto[0]) + "to" + str(datafromto[1]) + "list.html")

        fileext    = ".dat"

        filenolist = np.linspace(datafromto[0], datafromto[1], (datafromto[1]-datafromto[0]+1))

        #print(filenolist)
        scanno_list    = []
        command_list   = []
        scantitle_list = []

        basestr = "000000"

        for fileindex, fileno in enumerate (filenolist):
            filename = str(int(fileno))
            filename = basestr[:-len(filename)] + filename
            expnumstr = "0000"

            expnumstr=expnumstr[0:4-len(str(self.expnum))] + str(self.expnum)
                
            fullpath = ppath / Path(expnumstr + "_" + filename + fileext)
            #print(fullpath)

            scanno_list.append(int(fileno))
            with fullpath.open() as f:
                totallines = list(f)                              #read the lines in the file f, initialize a list
                for index, line in enumerate (totallines):

                    if (line.find("# command =") != -1):
                        command_list.append(line[11:-1])
                        #print(command_list)
                        #print(index)
                    if (line.find("# scan_title =")!= -1):
                        scantitle_list.append(line[14:-1])

        scan_dict ={"scanno": scanno_list, "command": command_list, "scantitle": scantitle_list}
        scanlist=pd.DataFrame(scan_dict)
        scanlist.to_html(str(outputfile))
        #print(scanlist)
        ### end of the function###
        return scanlist


    def sika_multichannel_config(self, total_channel_number=0, total_det_number=0, all_detname_list=None, all_det_eff_list=None, s2_offset_list=None, det_channel_list=None):
        if total_channel_number <3:
            total_channel_number = 9
        if total_det_number <3: 
            total_det_number = 48
        if all_detname_list is None:
            tubenumber=list(range(total_det_number))
            all_detname_list=[]
            for ii in tubenumber:
                all_detname_list.append("tube"+str(int(ii+1)))
            #print(all_detname_list)
        if all_det_eff_list is None:
            all_det_eff_list = [1.492784922, 0.679198979, 0.749288309, 0.916560322, 0.933051929, 0.700696967,
                                0.764798272, 1.005497202, 1.022577795, 0.675959556, 0.879257878, 0.954648081,
                                1.113968784, 0.59890056,  0.984490036, 0.839893983, 0.852851674, 0.734269167,
                                0.735250810, 0.991557868, 0.855207618, 0.744576421, 0.721409640, 1.000000000,
                                1.041229017, 0.779326593, 0.862569942, 0.926376755, 1.073328752, 0.654068911,
                                1.000588986, 0.945322470, 0.875134976, 0.822518897, 0.722685776, 0.982035928,
                                0.967605772, 0.826838127, 0.703936390, 0.948365564, 0.935898694, 0.726808678,
                                0.892019240, 0.928536370, 1.061058211, 0.711789536, 1.114459605, 1.278099539 ]
        if s2_offset_list is None:
            s2_offset_list = [-10, -7.5, -5.0, -2.50, 0, 2.5, 5.0, 7.5, 10.0]
        if det_channel_list is None:
            det_channel_list = [[7], [11], [15], [20], [24], [28], [32], [37], [41, 42]]

        super().tas_multichannel_config(total_channel_number, total_det_number,all_detname_list,all_det_eff_list,
        s2_offset_list,det_channel_list)
        return

    
    def sika_multichannel_reduction(self, path="", scanno=None):

        if self.multichannel == True:
            df = self.sika_data_to_pd(path, scanno)
            multichannel_dflist = super().tas_multichannel_reduction(df)
            return multichannel_dflist
        else:
            print("The multichannel has not been configured.")
            return


    def sika_multichannel_plot3d(self, path="", scanlist=None, xcol='qh', ycol='ql', zcol="en", perc_low=5, perc_high=95, xlim=None, ylim=None, zlim=None,xlabel="QH [r.l.u]", ylabel="QL [r.l.u]",zlabel="E [meV]"):
        """ put all the data points in dflist into X, Y, Z, Intensity, then plot in the 3d space with scatter. 
        """
        if scanlist is not None:
            alldflist=[]
            for scan in scanlist:
                df = self.sika_data_to_pd(path, scan)
                dflist = super().tas_multichannel_reduction(df)
                alldflist = alldflist + dflist
            X, Y, Z, A, ax= super().tas_multichannel_plot3d(alldflist,xcol='qh', ycol='ql', zcol="en", perc_low=5, perc_high=95, xlim=None, ylim=None, zlim=None,xlabel="QH [r.l.u]", ylabel="QL [r.l.u]",zlabel="E [meV]")
            return X, Y, Z, A, ax
        else:
            print("No scanlist is provided.")
            return


    def sika_expconfig(self, ef=5, aa=4, bb=5, cc=6, ang_a=90, ang_b=90, ang_c=90, uu=np.array([1, 0, 0]), vv=np.array([0, 0, 1])):
        sika_exp = super().tas_expconfig(ef)

        sika_exp.method   =   1  # 1 for Popovici, 0 for Cooper-Nathans
        sika_exp.moncor   =   1
        sika_exp.efixed   =   ef
        sika_exp.infin    =  -1    #const-Ef
        
        sika_exp.mono.dir =  1
        sika_exp.mono.dir = -1 
        sika_exp.ana.dir  =  1
        

        #Put the sample information below
        sika_exp.sample.a    = aa
        sika_exp.sample.b    = bb
        sika_exp.sample.c    = cc
        sika_exp.sample.alpha = ang_a
        sika_exp.sample.beta  = ang_b
        sika_exp.sample.gamma = ang_c
        sika_exp.sample.u       = uu
        sika_exp.sample.v       = vv
        sika_exp.hcol =  [60, 60, 60, 60]
        sika_exp.arms =  [230, 203, 179, 39.5, 106]
        sika_exp.orient1 = uu
        sika_exp.orient2 = vv                #need to be updated


        return sika_exp


    def sika_conv_init(self, hklw=None, exp = None,  initial=None, magion= "Mn2", sqw=SqwDemo, pref=PrefDemo):
        #numpy array is a dataframe and should be change into a numpy array
        if hklw is None:
            raise ValueError("hklw is None.") 
        if exp is None:
            exp = self.sika_expconfig(ef = 5.0)

        if initial is None:
            raise ValueError("initial is None.")

        [H, K, L, W, Iobs] =  hklw.to_numpy().T  #split into 1D arrays 
        #dIobs   = np.sqrt(Iobs)

        ffactor=SelFormFactor(magion)
        if ffactor is None:
            print("No Form Factor is used.")
        else:
            AA=ffactor["AA"]
            aa=ffactor["aa"]
            BB=ffactor["BB"]
            bb=ffactor["bb"]
            CC=ffactor["CC"]
            cc=ffactor["cc"]
            DD=ffactor["DD"]
            initial = list(initial) + [AA, aa, BB, bb, CC, cc, DD]  #new initial list
     

        newH=np.linspace(H[0], H[-1], 101)
        newK=np.linspace(K[0], K[-1], 101)
        newL=np.linspace(L[0], L[-1], 101)
        newW=np.linspace(W[0], W[-1], 101)     
        sim_init = exp.ResConv(sqw, pref, nargout=2, hkle=[newH,newK,newL,newW], METHOD='fix', ACCURACY=[5,5], p=initial)
        simHKLW=np.column_stack([newH,newK,newL,newW,sim_init])
        
        return simHKLW

    def sika_convfit(self, hklw=None, exp = None,  initial=None, fixlist=[0,0,0,0,0,0,0,0], magion = "none", sqw=SqwDemo, pref=PrefDemo):
        #numpy array is a dataframe and should be change into a numpy array
        if hklw is None:
            raise ValueError("hklw is None.") 
        if exp is None:
            exp = self.sika_expconfig(ef = 5)

        if initial is None:
            raise ValueError("initial is None.") 

        [H, K, L, W, Iobs] =  hklw.to_numpy().T  #split into columns
        dIobs=np.sqrt(Iobs)
        #load the FF parameters if magion is given
        ffactor=SelFormFactor(magion)
        if ffactor is None:
            print("No Form Factor is used.")
        else:
            AA=ffactor["AA"]
            aa=ffactor["aa"]
            BB=ffactor["BB"]
            bb=ffactor["bb"]
            CC=ffactor["CC"]
            cc=ffactor["cc"]
            DD=ffactor["DD"]
            initial = list(initial) + [AA, aa, BB, bb, CC, cc, DD]
            fixlist = fixlist       + [ 0,  0,  0,  0,  0,  0,  0]
            print(f"{magion} Form Factor is used.")

        fitter = UltraFastFitConv(exp, sqw, pref, [H,K,L,W], Iobs, dIobs)
        result = fitter.fit_ultrafast(param_initial=initial, param_fixed_mask=fixlist,maxfev=200,use_analytical_jacobian=True,early_stopping=True,verbose=True)
        
        final_params = result['params']
        # To produce a smooth fitted curve
        newH=np.linspace(H[0], H[-1], 101)
        newK=np.linspace(K[0], K[-1], 101)
        newL=np.linspace(L[0], L[-1], 101)
        newW=np.linspace(W[0], W[-1], 101)
        final      = exp.ResConv(sqw=sqw, pref=pref, nargout=2, hkle=[newH,newK,newL,newW], METHOD='fix', ACCURACY=None, p=final_params)
        fittedHKLW = np.column_stack([newH,newK,newL,newW,final])
        
        return result, fittedHKLW


    def sika_batchfile_validate(self, batchfile=None):
        # Now read and process the file
        if batchfile ==None:
            print("Please give the batch file!")
            return
        valid_cmd_list    = []
        validation_result = ""
        validator         = SikaCommandValidator()
        try:
            with open(batchfile, 'r') as file:
                lines = file.readlines()
                
            # Process lines
            i = 0
            while i < len(lines):
                # Skip empty or whitespace-only lines
                if not lines[i].strip():
                    i += 1
                    continue
                    
                current_line = lines[i].strip()
                print(f"\nProcessing line {i + 1}:")
                validation_result = validation_result+ f"\nLine {i + 1}:\n"
                print(f"Content: {current_line}")
                
                # Process mscan command
                valid, message = validator.validate_command(current_line)

                print(f"Command: {current_line}")
                
                
                if valid:
                    print(f"Validation passed!")
                    validation_result = validation_result+ f"Command: {current_line}"+"\n"+f"Validation passed"+f": {message}:\n\n"
                    valid_cmd_list.append(current_line)
                else: 
                    print(f"Error: Invalid")
                    validation_result = validation_result+ f"Command: {current_line}"+"\n"+f"Error: "+f"Message: {message}\n\n"
                i += 1
                    
        except FileNotFoundError:
            print("Error: Test file not found")
        except Exception as e:
            print(f"Error processing file: {str(e)}")

        return validation_result, valid_cmd_list

    def sika_batchcmd_validate(self, cmdlist=None):
        # Now read and process the file
        if cmdlist ==None:
            print("please give the batch files!")
            return
        valid_cmd_list = []
        validation_result = ""
        validator=SikaCommandValidator()
        try:

            i = 0
            for cmdline in cmdlist:
                # Skip empty or whitespace-only lines
                if not cmdline.strip():
                    i += 1
                    continue
                    
                current_line = cmdline.strip()
                print(f"\nProcessing line {i + 1}:")
                validation_result = validation_result+ f"\nLine {i + 1}:\n"
                print(f"Content: {current_line}")
                
                # Process mscan command
                valid, message = validator.validate_command(current_line)

                print(f"Command: {current_line}")
                
                
                if valid:
                    print(f"Validation passed!")
                    validation_result = validation_result+ f"Command: {current_line}"+"\n"+f"Validation passed"+f": {message}:\n\n"
                    valid_cmd_list.append(current_line)
                else: 
                    print(f"Error: Invalid")
                    validation_result = validation_result+ f"Command: {current_line}"+"\n"+f"Error: "+f"Message: {message}\n\n"
                i += 1
                    
        except FileNotFoundError:
            print("Error: Test file not found")
        except Exception as e:
            print(f"Error processing file: {str(e)}")

        return validation_result, valid_cmd_list
    

    def sika_scansim(self, command=None, exp=None):
        """
        Parse a scan command and return array of [h, k, l, e] positions
        Returns None if command is invalid
        """
        # First validate the command
        if not command:
            print("Please input the command!")
            return
        if not exp:
            print("Please input the exp!")
            return

        validator=SikaCommandValidator()

        # Initialize default values
        motor_values = {'h': None, 'k': None, 'l': None, 'e': None}
        scan_motor = None
        scan_range = None
        hpos = []
        kpos = []
        lpos = []
        epos = []
        n_steps    =   0
        last_steps = -99
        scan_motor=[]
        scan_range=[]
        fix_motor=[]
        fix_pos=[]

        df_scanpos = pd.DataFrame([])
        try:
            # Split command into parts
            parts = command.split()

            i = 1
            while i < len(parts):
                motor = parts[i].lower()
                if motor not in motor_values:
                    raise ValueError(f"Invalid motor: {motor}")
                    
                # Parse values
                if len(parts) > i + 3 and all(validator.is_number(parts[j]) for j in range(i+1, i+4)):
                    # This is the scanning motor (start, end, step)
                    scan_motor.append( motor)
                    start = float(parts[i+1])
                    end = float(parts[i+2])
                    step = float(parts[i+3])
                    scan_range.append([start, end, step])
                    if step == 0:
                        print("error: step size can not be zero.")
                        return None
                    
                    i += 4

                else:
                    # This is a fixed position
                    fix_motor.append(motor)
                    fix_pos.append(float(parts[i+1]))
                    #motor_values[motor] = float(parts[i+1])
                    i += 2
            
            if not scan_motor:
                raise ValueError("No scanning motor specified")
            #print(f"scan_motor:{scan_motor}")

            for idx, motor in enumerate(scan_motor):

                n_steps = abs(int(round((scan_range[idx][1] - scan_range[idx][0]) / scan_range[idx][2]))) + 1
                    
                if last_steps == -99 or last_steps == n_steps:
                    if motor == 'h':
                        hpos = np.linspace(scan_range[idx][0], scan_range[idx][1], n_steps)
                    elif motor == 'k':
                        kpos = np.linspace(scan_range[idx][0], scan_range[idx][1], n_steps)
                    elif motor == 'l':
                        lpos = np.linspace(scan_range[idx][0], scan_range[idx][1], n_steps)
                    elif motor == 'e':
                        epos = np.linspace(scan_range[idx][0], scan_range[idx][1], n_steps)
                    last_steps = n_steps
                else:
                    #print("last_steps:{}".format(last_steps))
                    #print("n_steps:{}".format(n_steps))
                    print("ERROR: the step numbers in the multi-motor scan do not match!")
                    return None
            
            for idx, motor in enumerate(fix_motor):
                if   motor == 'h':
                    hpos = np.linspace(fix_pos[idx], fix_pos[idx], n_steps)
                elif motor == 'k':
                    kpos = np.linspace(fix_pos[idx], fix_pos[idx], n_steps)
                elif motor == 'l':
                    lpos = np.linspace(fix_pos[idx], fix_pos[idx], n_steps)
                elif motor == 'e':
                    epos = np.linspace(fix_pos[idx], fix_pos[idx], n_steps)


            if len(hpos) >0:
                if len(hpos) == len(kpos)  and  len(hpos) == len(lpos)   and   len(hpos) == len(epos):    
                    hkle = np.stack([hpos, kpos,lpos,epos])  # doesn't need to transpose
                    #print(hkle)
                    [M1, M2, S1, S2, A1, A2, Q] = exp.get_spec_angles(hkle)

                    df_scanpos['h'] = hkle[0]
                    df_scanpos['k'] = hkle[1]
                    df_scanpos['l'] = hkle[2]
                    df_scanpos['e'] = hkle[3]
                    
                    df_scanpos['m1'] = M1
                    df_scanpos['m2'] = M2
                    df_scanpos['s1'] = S1
                    df_scanpos['s2'] = S2
                    df_scanpos['a1'] = A1
                    df_scanpos['a2'] = A2
                    #print(df_scanpos)
                    return df_scanpos
            else:
                print("There are some errors!")
                return  None
            
        except (ValueError, IndexError) as e:
            print(f"Error when parsing command: {str(e)}")
            return None



from pathlib import Path
import pandas as pd
import numpy as np
import math

import matplotlib.pyplot as plt
import mpl_toolkits.mplot3d as Axes3D
from matplotlib import cm
from matplotlib.colors import Normalize

from scipy.interpolate import griddata, make_interp_spline
from .utils.toolfunc import (ni_s2_residual, gaussian_residual, lorentzian_residual, 
                           fit_peak, angle2, SqwDemo, PrefDemo, SelFormFactor, 
                           PrefDemoQscan, strisfloat, AnglesToQhkl,set_standard_plot)

import inspy as npy
from inspy import TripleAxisSpectr
from inspy.insfit import FitConv, UltraFastFitConv


class TasData:
    """
    TAS (Triple Axis Spectrometer) Data analysis class.
    
    This class handles pandas DataFrames containing reduced TAS data with standardized
    column names including: 'qh', 'qk', 'ql', 'en', 'ei', 'ef', 'q', 'm1', 'm2', 
    's1', 's2', 'a1', 'a2', 'sgu', 'sgl', 'det', 'mon', 'time', 'sampleT1', 
    'sampleT2', 'magField'
    """
    # Constants
    PG002_D_SPACING = 3.355  # Angstroms
    #DEFAULT_NORM_COUNT = 1000000
    VARIANCE_THRESHOLD = 0.0006
    
    def __init__(self, tasname, expnum, title, sample, user):
        """
        Initialize TasData instance.
        
        Args:
            tasname (str): Name of the TAS instrument
            expnum (str or int): Experiment number
            title (str): Experiment title
            sample (str): Sample description
            user (list or tuple): List of users (non-empty)
        """
        if not all(isinstance(x, str) for x in [tasname, title, sample]): 
            raise ValueError("tasname, title, sample must be strings")
        if not isinstance(user, (list, tuple)) or not user: 
            raise ValueError("user must be a non-empty sequence")
        if not isinstance(expnum, (str, int)):
            raise ValueError("expnum must be a number or string")
            
        # Configuration
        self.tasname = tasname
        self.expnum = str(expnum)
        self.title = title
        self.sample = sample
        self.user = user
        self.exp = None
        self.multichannel = False
        self.multidetector_config = None
        self.multichannel_config = None
        self.total_channel_number = 1
        self.total_det_number = 1 

    def print_exp(self):
        """Print experiment information."""
        print(f'This is a TAS experiment {self.expnum} entitled with {self.title} '
              f'using {self.sample} by {self.user[0]} et al.')


    def tas_datanormalize(self, dflist=None, norm_mon_count=-999999):
        """
        Normalize detector counts by monitor counts.
        
        Args:
            dflist (list): List of pandas DataFrames to normalize
            norm_mon_count (int): Target monitor count for normalization
            
        Returns:
            list: List of normalized DataFrames
        """
        if not isinstance(dflist, list): 
            raise TypeError("dflist must be a list")
        if len(dflist[0]) == 0: 
            raise TypeError("the first dataframe in dflist is empty.")

        normalized_dflist = []

        # Determine norm_mon_count from first df if not provided
        if norm_mon_count < 0:
            first_df = dflist[0]
            first_df = first_df[first_df["monitor"] != 0]  # exclude zero-monitor rows before averaging
            norm_mon_count = first_df["monitor"].mean()

        for index, df in enumerate(dflist):
            if isinstance(df, pd.DataFrame):
                df = df.copy()                                              # avoid modifying original
                df['detector'] = df['detector'].astype(float)
                #df = df[df["monitor"] != 0]                                # remove zero-monitor rows
                # Only check the last row
                if df["monitor"].iloc[-1] < df["monitor"].mean() * 0.90:
                    df = df.iloc[:-1]
                # Skip empty dataframes after filtering
                if len(df) == 0:
                    normalized_dflist.append(df)
                    print(f"Warning: DataFrame at index {index} became empty after filtering")
                    continue

                if df["monitor"].iloc[-1] < df["monitor"].mean() * 0.90:  # incomplete last point
                    df = df.iloc[:-1]                                      # drop last row

                # Skip if dropping last row made it empty
                if len(df) == 0:
                    normalized_dflist.append(df)
                    print(f"Warning: DataFrame at index {index} became empty after dropping last row")
                    continue

                df['detector'] = df['detector'] * norm_mon_count / df['monitor']
                normalized_dflist.append(df)
            else:
                print(f"Error: element at index {index} in dflist is not a DataFrame.")

        return normalized_dflist


    def tas_datacombine(self, dflist=None, norm_mon_count=-999999):
        """
        Combine multiple scans into a single DataFrame.
        
        Args:
            dflist (list): List of DataFrames to combine
            norm_mon_count (int): Monitor count for normalization
            
        Returns:
            pd.DataFrame: Combined DataFrame
        """
        if not isinstance(dflist, list): 
            raise TypeError("dflist must be a list")
            
        if len(dflist) == 1:
            print("Only one dataset, no need to combine.")
            return dflist[0]
            
        # Normalize first
        dflist = self.tas_datanormalize(dflist, norm_mon_count)
        comb_df = pd.DataFrame()

        Xaxis = ""
        for index, df in enumerate(dflist):
            if isinstance(df, pd.DataFrame):                
                curXaxis = df.attrs.get('scanax1')
                if curXaxis is None:
                    raise TypeError("The scan axis is unknown from the data header.")
                
                if index == 0:
                    Xaxis = curXaxis
                    comb_df = df.copy()
                    if 'scanax1' not in df.attrs: 
                        raise ValueError("DataFrame missing scanax1 attribute")
                    comb_df.attrs = df.attrs.copy()

                elif curXaxis == Xaxis:
                    for jj in df.index:
                        point_exist = False
                        for nn in comb_df.index:
                            # Check if positions are close enough to be considered the same
                            if np.abs(df[Xaxis].iloc[jj] - comb_df[Xaxis].iloc[nn]) < 0.001:
                                point_exist = True
                                # Average the detector counts
                                comb_df.loc[nn, 'detector'] = (df.loc[jj, 'detector'] + 
                                                              comb_df.loc[nn, 'detector']) / 2
                        
                        if not point_exist:
                            current_row = df.iloc[jj].to_dict()
                            new_index = len(comb_df)
                            for col, val in current_row.items():
                                comb_df.loc[new_index, col] = val
                    
                    # Update scan number attribute
                    comb_df.attrs['scanno'] = (comb_df.attrs.get('scanno') + 
                                             "+" + df.attrs.get('scanno'))
                else:
                    print(f"The scan #{index} axis is not the same as the first scan!")
        
        if not comb_df.empty:
            comb_df = comb_df.sort_values(by=[Xaxis])
            comb_df.reset_index(drop=True, inplace=True)

        return comb_df

    def tas_export_hklw(self, dflist=None, hklw_file=""):
        """
        Export HKLW data to CSV file.
        
        Args:
            dflist (list): List of DataFrames to export
            hklw_file (str): Output filename
            
        Returns:
            pd.DataFrame: Exported HKLW data
        """
        if not isinstance(dflist, list): 
            raise TypeError("dflist must be a list")
    
        hklw = None
        for df in dflist:
            required_cols = ['qh', 'qk', 'ql', 'en', 'detector']
            if not all(col in df.columns for col in required_cols): 
                raise ValueError(f"Missing required columns: {required_cols}")
                
            hklw = df[required_cols].copy()
            scanno = df.attrs.get('scanno')
            
            if hklw_file == "":
                hklw_file = f"{self.tasname}_exp{self.expnum}_HKLW_#{scanno}.csv"
            hklw.to_csv(Path(hklw_file), index=False)

        return hklw

    def tas_simpleplot(self, dflist=None, fit=False, initial=None):
        """
        Simple plot and fit of TAS data.
        
        Args:
            dflist (list): List of DataFrames to plot
            fit (bool): Whether to perform fitting
            initial (dict): Initial parameters for fitting
            
        Returns:
            tuple: (fitted_df, params_df, fig, axs)
        """
        if not dflist or not all(isinstance(df, pd.DataFrame) for df in dflist): 
            raise ValueError("dflist must be a non-empty list of DataFrames")
        
        params_df  =  pd.DataFrame(columns = ['A', 'A_err', 'w','w_err', 'x0', 'x0_err', 'bg', 'bg_err'])
        fitted_df  =  pd.DataFrame([])
        Xaxis      = "" 
        Yaxis      = "detector"

        if len(dflist) > 1:
            fig, axs = plt.subplots(len(dflist), 1, figsize=(6, 4*len(dflist)))
            # Convert axs to list for consistency
            axs = list(axs)    # NEW: Ensure axs is a list
                
            for ii in range(len(dflist)):
                if dflist[ii].empty:
                    print(f"The {ii}th DataFrame in the list is empty! No plot.")
                else:
                    Xaxis = dflist[ii].attrs.get('scanax1')
                    if Xaxis is None:
                        raise TypeError("The scan axis is unknown from the data header.")

                    dataX = dflist[ii][Xaxis].to_numpy()
                    dataY = dflist[ii][Yaxis].to_numpy()
                    
                    if fit and len(dflist[ii]) > 4:
                        cur_fitpar, cur_fitdat = fit_peak(dataX, dataY, func='G', initial=initial)
                        axs[ii].plot(dataX, dataY, 'o-', cur_fitdat["X"], cur_fitdat["Y_fit"], '-')

                        params_df   =  pd.concat([params_df, cur_fitpar], axis=0, ignore_index=True)
                        fitted_df   =  pd.merge(fitted_df, cur_fitdat, how='outer', left_index=True, right_index=True, suffixes=("", "_"+str(ii)))
                    else:
                        axs[ii].plot(dataX, dataY, 'o-')
                        if fit and len(dflist[ii]) <= 4:
                            print("Not enough data points for fitting!")
                    axs[ii].set_xlabel(Xaxis)
                    axs[ii].set_ylabel('Intensity [counts]')

            return params_df, fitted_df, fig, axs

        elif len(dflist) == 1:
            if dflist[0].empty:
                print(f"The only one DataFrame in the list is empty! No plot.")
            else:
                fig, ax = plt.subplots(1, 1, figsize=(6, 4))
                Xaxis = dflist[0].attrs.get('scanax1')
                if Xaxis is None:
                    raise TypeError("The scan axis is unknown from the data header.")

                dataX = dflist[0][Xaxis].to_numpy()
                dataY = dflist[0][Yaxis].to_numpy()
        
                if fit and len(dflist[0]) > 4:
                    cur_fitpar, cur_fitdat = fit_peak(dataX, dataY, func='G', initial=initial)
                    ax.plot(dataX, dataY, 'o-', cur_fitdat["X"], cur_fitdat["Y_fit"], '-')

                    params_df   =  pd.concat([params_df, cur_fitpar], axis=0, ignore_index=True)
                    fitted_df   =  pd.merge(fitted_df, cur_fitdat, how='outer', left_index=True, right_index=True, suffixes=("", "_"))
                else:
                    ax.plot(dataX, dataY, 'o-')
                    if fit and len(dflist[0]) <= 4 :
                        print("Not enough data points for fitting!")

                ax.set_xlabel(Xaxis)
                ax.set_ylabel('Intensity [counts]')

            return params_df, fitted_df, fig, [ax]  # NEW: Return ax as a single-item list

 
    def tas_combplot(self, dflist=None, fit=False, norm_mon_count = -999999, overplot=False, offset=1000, initial=None):
        """
        Combined plot of multiple TAS scans.
        
        Args:
            dflist (list): List of DataFrames or lists of DataFrames to plot
            fit (bool): Whether to perform fitting
            norm_mon_count (int): Monitor count for normalization
            overplot (bool): Whether to plot all scans on same axes
            offset (float): Y-offset for overplotted scans
            initial (dict): Initial parameters for fitting
            
        Returns:
            tuple: (fitted_df, params_df, fig, axs) where axs is always a list of Axes
        """
        if not dflist or not all(isinstance(df, (pd.DataFrame, list)) for df in dflist): 
            raise ValueError("dflist must be a non-empty list of DataFrames or lists")

        params_df  =  pd.DataFrame(columns = ['A', 'A_err', 'w','w_err', 'x0', 'x0_err', 'bg', 'bg_err'])
        fitted_df  =  pd.DataFrame([])
        Xaxis      =  ""
        Yaxis      = "detector"
        #print(dflist)
        dflist = self.tas_datanormalize(dflist, norm_mon_count)
        #print(dflist)
        if not overplot and len(dflist) > 1:
            fig, axs = plt.subplots(len(dflist), 1, figsize=(6, 4*len(dflist)))
            # Convert axs to list for consistency (in case it's a NumPy array)
            axs = list(axs)

            for ii, scan in enumerate(dflist):
                if isinstance(scan, list):
                    print("problem in tas_combplot")
                    hklw = self.tas_datacombine(scan, norm_mon_count)
                    Xaxis = hklw.attrs.get('scanax1')
                    if Xaxis is None:
                        raise TypeError("The scan axis is unknown from the data header.")
                else:
                    hklw = scan.copy()
                    hklw.attrs = scan.attrs.copy()
                    Xaxis = hklw.attrs.get('scanax1')
                    if Xaxis is None:
                        raise TypeError("The scan axis is unknown from the data header.")
                    
                dataX = hklw[Xaxis].to_numpy()
                dataY = hklw[Yaxis].to_numpy()

                if fit and len(hklw) > 4:
                    cur_fitpar, cur_fitdat = fit_peak(dataX, dataY, func='G', initial=initial)
                    axs[ii].plot(dataX, dataY, 'o-', cur_fitdat["X"], cur_fitdat["Y_fit"], '-')
                    params_df   =  pd.concat([params_df, cur_fitpar], axis=0, ignore_index=True)
                    fitted_df   =  pd.merge(fitted_df, cur_fitdat, how='outer', left_index=True, right_index=True, suffixes=("", "_"+str(ii)))
                else:
                    axs[ii].plot(dataX, dataY, 'o-')
                    if fit and len(hklw) < 4 :
                        print("Not enough data points for fitting!")

                axs[ii].set_xlabel(Xaxis)
                axs[ii].set_ylabel('Intensity [counts]')
            
            return params_df, fitted_df, fig, axs
        else:
            # overplot=True OR single scan case
            fig, ax = plt.subplots(1, 1, figsize=(6, 8))
            
            # Loop through all scans in dflist
            for ii, scan in enumerate(dflist):
                if isinstance(scan, list):

                    hklw = self.tas_datacombine(scan, norm_mon_count)
                    Xaxis = hklw.attrs.get('scanax1')
                    if Xaxis is None:
                        raise TypeError("The scan axis is unknown from the data header.")
                else:
                    hklw = scan.copy()
                    hklw.attrs = scan.attrs.copy()
                    Xaxis = hklw.attrs.get('scanax1')
                    if Xaxis is None:
                        raise TypeError("The scan axis is unknown from the data header.")
                
                dataX = hklw[Xaxis].to_numpy()
                dataY = hklw[Yaxis].to_numpy()
                
                # Apply offset when overplotting multiple scans
                if overplot and ii > 0:
                    dataY = dataY + offset * ii
                
                if fit and len(hklw) > 4:
                    cur_fitpar, cur_fitdat = fit_peak(dataX, dataY, func='G', initial=initial)
                    ax.plot(dataX, dataY, 'o-', cur_fitdat["X"], cur_fitdat["Y_fit"], '-')
    
                    params_df   =  pd.concat([params_df, cur_fitpar], axis=0, ignore_index=True)
                    fitted_df   =  pd.merge(fitted_df, cur_fitdat, how='outer', left_index=True, right_index=True, suffixes=("", "_"+str(ii)))
                else:
                    ax.plot(dataX, dataY, 'o-')
                    if fit and len(hklw) < 4:
                        print("Not enough data points for fitting!")

            ax.set_xlabel(Xaxis)
            ax.set_ylabel('Intensity [counts]')

            return params_df, fitted_df, fig, [ax]

    def tas_combspline(self, dflist=None, fit=False, norm_mon_count = -999999, overplot=False, offset=1000, initial=None):
        """
        Combined plot of multiple TAS scans.
        
        Args:
            dflist (list): List of DataFrames or lists of DataFrames to plot
            fit (bool): Whether to perform fitting
            norm_mon_count (int): Monitor count for normalization
            overplot (bool): Whether to plot all scans on same axes
            offset (float): Y-offset for overplotted scans
            initial (dict): Initial parameters for fitting
            
        Returns:
            tuple: (fitted_df, params_df, fig, axs) where axs is always a list of Axes
        """
        if not dflist or not all(isinstance(df, (pd.DataFrame, list)) for df in dflist): 
            raise ValueError("dflist must be a non-empty list of DataFrames or lists")
        colors = ['C0','C1','C2','C3','C4','C5','C6','C7','C8','C9','C10','C11','C12','C13','C14']
        params_df  =  pd.DataFrame(columns = ['A', 'A_err', 'w','w_err', 'x0', 'x0_err', 'bg', 'bg_err'])
        fitted_df  =  pd.DataFrame([])
        Xaxis      =  ""
        Yaxis      = "detector"
        #print(dflist)
        dflist = self.tas_datanormalize(dflist, norm_mon_count)
        #print(dflist)
        if not overplot and len(dflist) > 1:
            fig, axs = plt.subplots(len(dflist), 1, figsize=(6, 4*len(dflist)))
            # Convert axs to list for consistency (in case it's a NumPy array)
            axs = list(axs)

            for ii, scan in enumerate(dflist):
                if isinstance(scan, list):
                    print("problem in tas_combplot")
                    hklw = self.tas_datacombine(scan, norm_mon_count)
                    Xaxis = hklw.attrs.get('scanax1')
                    if Xaxis is None:
                        raise TypeError("The scan axis is unknown from the data header.")
                else:
                    hklw = scan.copy()
                    hklw.attrs = scan.attrs.copy()
                    Xaxis = hklw.attrs.get('scanax1')
                    if Xaxis is None:
                        raise TypeError("The scan axis is unknown from the data header.")
                    
                    # Sort by scanax to ensure ascending order for spline interpolation
                df_sorted = hklw.sort_values(by=Xaxis)
                dataX = df_sorted[Xaxis].to_numpy()
                dataY = df_sorted[Yaxis].to_numpy()
                    
                smoothX = np.linspace(dataX.min(), dataX.max(), len(df_sorted)*5+1)
                spl = make_interp_spline(dataX, dataY, k=3)
                smoothY = spl(smoothX)

                if fit and len(dataX) > 4:
                    cur_fitpar, cur_fitdat = fit_peak(dataX, dataY, func='G', initial=initial)
                    axs[ii].plot(dataX, dataY, 'o', color=colors[ii%15])
                    axs[ii].plot(smoothX, smoothY, '-', color=colors[ii%15] )
                    axs[ii].plot(cur_fitdat["X"], cur_fitdat["Y_fit"], '--', color=colors[ii%15])
                    params_df   =  pd.concat([params_df, cur_fitpar], axis=0, ignore_index=True)
                    fitted_df   =  pd.merge(fitted_df, cur_fitdat, how='outer', left_index=True, right_index=True, suffixes=("", "_"+str(ii)))
                else:
                    axs[ii].plot(dataX, dataY, 'o')
                    axs[ii].plot(smoothX, smoothY, '-', color=colors[ii%15])
                    if fit and len(dataX) < 4 :
                        print("Not enough data points for fitting!")

                axs[ii].set_xlabel(Xaxis)
                axs[ii].set_ylabel('Intensity [counts]')
            
            return params_df, fitted_df, fig, axs
        else:
            # overplot=True OR single scan case
            fig, ax = plt.subplots(1, 1, figsize=(6, 8))
            
            # Loop through all scans in dflist
            for ii, scan in enumerate(dflist):
                if isinstance(scan, list):

                    hklw = self.tas_datacombine(scan, norm_mon_count)
                    Xaxis = hklw.attrs.get('scanax1')
                    if Xaxis is None:
                        raise TypeError("The scan axis is unknown from the data header.")
                else:
                    hklw = scan.copy()
                    hklw.attrs = scan.attrs.copy()
                    Xaxis = hklw.attrs.get('scanax1')
                    if Xaxis is None:
                        raise TypeError("The scan axis is unknown from the data header.")
                

                df_sorted = hklw.sort_values(by=Xaxis)
                dataX = df_sorted[Xaxis].to_numpy()
                dataY = df_sorted[Yaxis].to_numpy()
                    
                smoothX = np.linspace(dataX.min(), dataX.max(), len(df_sorted)*5+1)
                spl = make_interp_spline(dataX, dataY, k=3)
                smoothY = spl(smoothX)
                
                # Apply offset when overplotting multiple scans
                if overplot and ii > 0:
                    dataY = dataY + offset * ii
                    smoothY = smoothY + offset * ii
                
                if fit and len(dataX) > 4:
                    cur_fitpar, cur_fitdat = fit_peak(dataX, dataY, func='G', initial=initial)
                    ax.plot(dataX, dataY, 'o', color=colors[ii%15])
                    ax.plot(smoothX, smoothY, '-', color=colors[ii%15])
                    ax.plot(cur_fitdat["X"], cur_fitdat["Y_fit"], '--', color=colors[ii%15])
    
                    params_df   =  pd.concat([params_df, cur_fitpar], axis=0, ignore_index=True)
                    fitted_df   =  pd.merge(fitted_df, cur_fitdat, how='outer', left_index=True, right_index=True, suffixes=("", "_"+str(ii)))
                else:
                    ax.plot(dataX, dataY, 'o')
                    ax.plot(smoothX, smoothY, '-', color=colors[ii%15])
                    if fit and len(dataX) < 4:
                        print("Not enough data points for fitting!")

            ax.set_xlabel(Xaxis)
            ax.set_ylabel('Intensity [counts]')

            return params_df, fitted_df, fig, [ax]


    def tas_random_contour(self, df_reduced, x_col="qh", y_col="en", xlabel='Q [rlu]', ylabel='Energy transfer [meV]',title=None, vminmax=[0, 1000], output_file=None):
        """
        Create a contour plot using matplotlib.
        
        Args:
            df_reduced (pd.DataFrame): DataFrame with reduced data
            x_col (str): X-axis column name
            y_col (str): Y-axis column name
            xtitle (str): X-axis label
            ytitle (str): Y-axis label
            vminmax (list): Min and max values for intensity
            output_file (str): Optional output filename
            
        Returns:
            matplotlib.axes.Axes: The plot axes
        """
        required_cols = [x_col, y_col, 'detector']
        if not all(col in df_reduced.columns for col in required_cols): 
            raise ValueError(f"Missing required columns: {required_cols}")

        x = df_reduced[x_col].values
        y = df_reduced[y_col].values
        z = df_reduced['detector'].values
        
        if len(x) < 3: 
            raise ValueError("Insufficient points for contour (need at least 3)")

        # Create regular grid for contour plot
        xi = np.linspace(min(x), max(x), 100)
        yi = np.linspace(min(y), max(y), 100)
        xi, yi = np.meshgrid(xi, yi)
        
        # Interpolate z values on the regular grid
        zi = griddata((x, y), z, (xi, yi), method='linear')
        zi = np.clip(zi, vminmax[0], vminmax[1])  # Clip values to defined range
        zi = np.nan_to_num(zi, nan=vminmax[0])  # Replace NaN with min value
        
        # Create the contour plot
        fig, ax = plt.subplots(figsize=(8, 6))
        cs = ax.contourf(xi, yi, zi, levels=100, vmin=vminmax[0], vmax=vminmax[1], cmap='viridis')

        set_standard_plot(ax, title, xlabel, ylabel, bGrid=True)
        fig.colorbar(cs, ax=ax)

        if output_file:
            plt.savefig(output_file, dpi=300, bbox_inches='tight')

        return xi, yi, zi, ax



    def tas_tidy_contour(self, dflist=None, x_col='qh', y_col='en', xlabel='Q [rlu]',ylabel='E [meV]', title=None,  vminmax=None, bRot=False, ax=None):
        """
        Combine scans into a contour plot.
        
        Args:
            dflist (list): List of DataFrames
            vminmax (list): Min/max values for contour
            x_col (str): X-axis column name
            y_col (str): Y-axis column name
            scan_range (list): Range for x-axis
            xlabel (str): X-axis label
            ylabel (str): Y-axis label
            title (str): Plot title
            ax (matplotlib.axes.Axes): Optional axes to plot on
            
        Returns:
            tuple: (xx, yy, intensity, ax) - grid coordinates, intensity, and axes
        """
        if not dflist or not all(isinstance(df, pd.DataFrame) for df in dflist): 
            raise ValueError("dflist must be non-empty list of DataFrames")

        if vminmax is None:
            vminmax = [0, 1000]

        contourdata =  pd.DataFrame([])

        scan_range  = []
        pos_range   = []
        Xaxis       = ""
        delta0      = 0


        if ax is None:
            fig, ax = plt.subplots(figsize=(6,6))

        for scan in dflist:
            Xaxis = scan.attrs['scanax1']   
            if Xaxis != y_col:
                raise ValueError("The scan axis of all data must be the same as y_col.")
        
        low_lim = min(df[y_col].min() for df in dflist) #use the min as the low range
        up_lim  = max(df[y_col].max() for df in dflist)  #use the max as the up range
        #print(f"lowlim:{low_lim}; uplim:{up_lim}")

        for scanindex, scan in enumerate (dflist):
                
            hklw       = scan[['qh','qk','ql','en','detector','monitor']]                #only choose the most important data
            hklw.attrs = scan.attrs.copy()
            datalines  = len(hklw.index)
            Xaxis = hklw.attrs['scanax1']   
            #print(f"the scanning axis is {Xaxis}") 

            if(scanindex==0):
                #get stepsize from the first data
                delta0      = np.abs(hklw[Xaxis][0]-hklw[Xaxis][datalines-1])/(datalines-1)    #determine the step size
                points      = int(np.around((up_lim-low_lim)/delta0)) +1                       #determine how many steps in whole range
                scan_range  = np.linspace(low_lim, up_lim, points)                             #generate an array with step size of delta
                contourx    = pd.DataFrame(data=scan_range)                                    #create a DataFrame with first column scan_range
                contourx.columns=[Xaxis]                                                       #give a name to this column as the scan axis
                contourdata = pd.merge(contourdata, contourx, how='outer', left_index=True, right_index=True)
            else:
                delta    = np.abs(hklw[Xaxis][0]-hklw[Xaxis][datalines-1])/(datalines-1)
                if np.abs(delta-delta0) > 0.002:
                    print("Error: the step size of the scan no. {} is not the same!".format(scanindex))
                    return 

            curCol       = hklw[Xaxis].to_numpy()                                                # get the range of the real data from file
            curColCount  = hklw['detector'].to_numpy()                                           # not necessary, normalization has done.
            pos_range.append(hklw[x_col][0])                                                     # generate an array along energy  y_col 

            step_deci, step_int = math.modf((curCol[datalines-1]-scan_range[0])/delta0)

            if np.abs(step_deci) > 0.45 and np.abs(step_deci) < 0.55:
                print(f"Warning: Scan #{hklw.attrs.get('scanno')}: half step detected!")
                curCol = curCol + delta0/2                                        # shift half step for all
                 
            steps        = int(np.around((curCol[datalines-1]-scan_range[0])/delta0))                #calculate the difference between max range and the last data point
            if (points-steps-1) > 0:
                temp = np.zeros(points-steps-1) + 0.1                        #generate a zero array of the size of the end
                curColCount = np.insert(curColCount, datalines, temp)        #insert the zero array to the end of the real data

            steps        = int(np.around((curCol[0]-scan_range[0])/delta0)) #calculate the missing points at the beginning of real data

            if steps > 0:
                temp = np.zeros(steps)  + 0.1                                  #generate a zero arry of size of the beginning
                curColCount = np.insert(curColCount, 0, temp)                  #insert it

            tidyarray = np.vstack((scan_range, curColCount)).T                      # now it has the same size as the low and high range
            tempdf    =  pd.DataFrame(tidyarray, columns=[Xaxis,"count_{}".format(scanindex)])
            
            contourdata = pd.merge(contourdata, tempdf, on=Xaxis, how='outer')  


        Yscan       = np.linspace(low_lim, up_lim, points)
        Xpos        = np.linspace(pos_range[0], pos_range[-1], len(pos_range) )
        
        intensity   = contourdata.drop(columns=[Xaxis]).to_numpy()

        if  bRot == False:
            cs  = ax.contourf(Xpos, Yscan, intensity, levels=900,  vmin=vminmax[0], vmax=vminmax[1]) #
        else:
            cs  = ax.contourf(Yscan, Xpos, intensity.T, levels=900,  vmin=vminmax[0], vmax=vminmax[1]) #


        set_standard_plot(ax, title, xlabel, ylabel)
        
        return Xpos, Yscan, intensity, ax



    
    def tas_bosecorrect(self,dflist=None, sample_T=1.5):
        """
        Apply Bose-Einstein population correction to inelastic neutron scattering data.
        
        Args:
            path (str): Directory path containing scan files
            dflist: List of scan numbers to process
            sample_T (float): Sample temperature in Kelvin
            
        Returns:
            list: List of corrected DataFrames
        """
        KB_MEV_K = 0.08617  # Boltzmann constant in meV/K (more accurate value)
        
        if sample_T <= 0:
            print('The sample temperature is incorrect, it should be given in Kelvin')
            return
        
        newdflist = []
        
        for index, df in enumerate(dflist):
            # Keep only positive energy transfers
            newdf = df.loc[df['en'] > 0].copy()  # Use .copy() to avoid SettingWithCopyWarning 
            newdf.reset_index(drop=True, inplace=True)  
            
            if df.attrs['scanax1'] == 'en':
                deltae = newdf['en'].to_numpy()
                counts = newdf['detector'].to_numpy()
                # Apply Bose population correction
                # Standard formula: I_corrected = I_measured / [1 - exp(-E/kT)]
                bose_factor = 1 - np.exp(-deltae / (sample_T * KB_MEV_K))     #from claude 
                # Avoid division by zero for very small energies
                bose_factor = np.where(np.abs(bose_factor) < 1e-10, 1e-10, bose_factor)
                counts_corrected = counts / bose_factor
                
                newdf['detector'] = counts_corrected
            else:
                print(f"Warning: Scan {index} is not an energy scan, skipping correction.")
            
            newdflist.append(newdf)
        
        return newdflist

    def tas_expconfig(self, ef=5):
        """
        Configure TAS experiment parameters.
        
        Args:
            ef (float): Fixed energy (default 5 meV)
            
        Returns:
            TripleAxisSpectr: Configured experiment object
        """
        tas_exp = TripleAxisSpectr(efixed=ef)
        tas_exp.method = 1  # 1 for Popovici, 0 for Cooper-Nathans
        tas_exp.moncor = 1
        tas_exp.efixed = ef
        tas_exp.infin = -1    # const-Ef
        
        tas_exp.mono.dir = 1
        tas_exp.ana.dir = 1
        
        # Monochromator configuration
        tas_exp.mono.tau = 'PG(002)'
        tas_exp.mono.mosaic = 30
        tas_exp.mono.vmosaic = 30
        tas_exp.mono.height = 10
        tas_exp.mono.width = 10
        tas_exp.mono.depth = 0.2
        tas_exp.mono.rh = 100
        tas_exp.mono.rv = 100
        
        # Analyzer configuration
        tas_exp.ana.tau = 'PG(002)'
        tas_exp.ana.mosaic = 30
        tas_exp.ana.vmosaic = 30
        tas_exp.ana.height = 10
        tas_exp.ana.width = 10
        tas_exp.ana.depth = 0.2
        tas_exp.ana.rh = 100
        tas_exp.ana.rv = 100
        
        # Sample information
        tas_exp.sample.a = 4
        tas_exp.sample.b = 5
        tas_exp.sample.c = 6
        tas_exp.sample.alpha = 90
        tas_exp.sample.beta = 90
        tas_exp.sample.gamma = 90
        tas_exp.sample.mosaic = 30
        tas_exp.sample.vmosaic = 30
        tas_exp.sample.u = np.array([1, 0, 0])
        tas_exp.sample.v = np.array([0, 0, 1])
        tas_exp.sample.shape_type = 'rectangular'
        tas_exp.sample.shape = np.diag([0.6, 0.6, 10]) ** 2
        
        # Collimation and arms
        tas_exp.hcol = [60, 60, 60, 60]
        tas_exp.vcol = [500, 500, 500, 500]
        tas_exp.arms = [200, 200, 200, 100, 100]
        tas_exp.orient1 = np.array([1, 0, 0])
        tas_exp.orient2 = np.array([0, 0, 1])
        
        # Guide and detector
        tas_exp.guide.height = 15
        tas_exp.guide.width = 5
        tas_exp.detector.height = 15
        tas_exp.detector.width = 2.5
        
        return tas_exp

    def tas_conv_init(self, hklw=None, exp=None, initial=None, magion="none", sqw=SqwDemo, pref=PrefDemo):
        """
        Initialize TAS convolution calculation.
        
        Args:
            hklw (pd.DataFrame): HKLW data
            exp (TripleAxisSpectr): Experiment configuration
            initial (dict): Initial parameters
            fixlist (list): Fixed parameter flags
            magion (str): Magnetic ion type
            sqw (function): Scattering function
            pref (function): Prefactor function
            smoothfit (bool): Whether to use smooth fitting
            
        Returns:
            np.ndarray: Initial simulation result
        """
        if hklw is None:
            raise ValueError("hklw is None.") 
        if exp is None:
            exp = self.taipan_expconfig(ef = 14.87)

        if initial is None:
            raise ValueError("initial is None.")

        [H, K, L, W, Iobs] =  hklw.to_numpy().T  #split into 1D arrays 
        #dIobs   = np.sqrt(Iobs)

        ffactor=SelFormFactor(magion)
        if ffactor is None:
            print("No Form Factor is used.")
        else:
            AA=ffactor["AA"]
            aa=ffactor["aa"]
            BB=ffactor["BB"]
            bb=ffactor["bb"]
            CC=ffactor["CC"]
            cc=ffactor["cc"]
            DD=ffactor["DD"]
            initial = list(initial) + [AA, aa, BB, bb, CC, cc, DD]  #new initial list
            print(f" {magion} Form Factor is used.")

        newH=np.linspace(H[0], H[-1], 101)
        newK=np.linspace(K[0], K[-1], 101)
        newL=np.linspace(L[0], L[-1], 101)
        newW=np.linspace(W[0], W[-1], 101)     
        sim_init = exp.ResConv(sqw, pref, nargout=2, hkle=[newH,newK,newL,newW], METHOD='fix', ACCURACY=[5,5], p=initial)
        simHKLW  = np.column_stack([newH,newK,newL,newW,sim_init])

        return simHKLW

    def tas_convfit(self, hklw=None, exp=None, initial=None, fixlist=[0,0,0,0,0,0,0,0], magion="none", sqw=SqwDemo, pref=PrefDemo):
        """
        Perform TAS convolution fitting.
        
        Args:
            hklw (pd.DataFrame): HKLW data
            exp (TripleAxisSpectr): Experiment configuration
            initial (dict): Initial parameters
            fixlist (list): Fixed parameter flags
            magion (str): Magnetic ion type
            sqw (function): Scattering function
            pref (function): Prefactor function
            smoothfit (bool): Whether to use smooth fitting
            
        Returns:
            tuple: (param_err, hkle_fit) - fitted parameters and smooth curve
        """
        if hklw is None:
            raise ValueError("hklw is None.") 
        if exp is None:
            exp = self.taipan_expconfig(ef = 14.87)

        if initial is None:
            raise ValueError("initial is None.") 

        [H, K, L, W, Iobs] =  hklw.to_numpy().T  #split into columns
        dIobs=np.sqrt(Iobs)
        #load the FF parameters if magion is given
        ffactor=SelFormFactor(magion)
        if ffactor is None:
            print("No Form Factor is used.")
        else:
            AA=ffactor["AA"]
            aa=ffactor["aa"]
            BB=ffactor["BB"]
            bb=ffactor["bb"]
            CC=ffactor["CC"]
            cc=ffactor["cc"]
            DD=ffactor["DD"]
            initial = list(initial) + [AA, aa, BB, bb, CC, cc, DD]
            fixlist = fixlist       + [ 0,  0,  0,  0,  0,  0,  0]
            print(f" {magion} Form Factor is used.")

        fitter = UltraFastFitConv(exp, sqw, pref, [H,K,L,W], Iobs, dIobs)
        result = fitter.fit_ultrafast(param_initial=initial, param_fixed_mask=fixlist,maxfev=200,use_analytical_jacobian=True,early_stopping=True,verbose=True)
        
        final_params = result['params']
        # To produce a smooth fitted curve
        newH=np.linspace(H[0], H[-1], 101)
        newK=np.linspace(K[0], K[-1], 101)
        newL=np.linspace(L[0], L[-1], 101)
        newW=np.linspace(W[0], W[-1], 101)
        final = exp.ResConv(sqw, pref, nargout=2, hkle=[newH,newK,newL,newW], METHOD='fix', ACCURACY=None, p=final_params)
        fittedHKLW=np.column_stack([newH,newK,newL,newW,final])
        
        return result, fittedHKLW

    def tas_multichannel_config(self, total_channel_number=21, total_det_number=48, 
                               all_detname_list=None, all_det_eff_list=None, 
                               s2_offset_list=None, det_channel_list=None):
        """
        Configure multichannel detector setup.
        
        Args:
            total_channel_number (int): Total number of channels
            total_det_number (int): Total number of detector tubes
            all_detname_list (list): List of detector names
            all_det_eff_list (list): List of detector efficiencies
            s2_offset_list (list): List of S2 angle offsets for each channel
            det_channel_list (list): List of detector-to-channel mappings
        """
        if not all(isinstance(x, (list, tuple)) and x is not None 
                  for x in [all_detname_list, all_det_eff_list, s2_offset_list, det_channel_list]): 
            raise ValueError("All list inputs must be non-None lists")
            
        if (len(all_detname_list) != total_det_number or 
            len(all_det_eff_list) != total_det_number or
            len(s2_offset_list) != total_channel_number or 
            len(det_channel_list) != total_channel_number): 
            raise ValueError("List lengths mismatch with specified counts")

        self.total_channel_number = total_channel_number
        self.total_det_number = total_det_number 

        # Set up detector configuration
        self.multidetector_config = pd.DataFrame({
            "all_detname": all_detname_list, 
            "all_det_eff": all_det_eff_list
        })
        
        # Set up channel configuration
        self.multichannel_config = pd.DataFrame({
            "s2_offset": s2_offset_list, 
            "det_channel": det_channel_list
        })
        
        self.multichannel = True
        print("The multichannel configuration setup is completed.")
        return

    def tas_multichannel_reduction(self, df=None):
        """
        Reduce multichannel detector data.
        
        Args:
            df (pd.DataFrame): Input DataFrame with detector data
            
        Returns:
            list: List of DataFrames for each channel
        """
        if df is None or df.empty: 
            raise ValueError("DataFrame is required and cannot be empty")

        if 'ubmatrix' not in df.attrs: 
            raise ValueError("ubmatrix missing in df.attrs")

        if not self.multichannel:
            print("The multichannel has not been configured.")
            return None

        multichannel_dflist = []
        UBMat = np.matrix(df.attrs['ubmatrix']).reshape(3, 3)

        multi_detname_list = self.multidetector_config['all_detname'].tolist()
        multi_deteff_list = self.multidetector_config['all_det_eff'].tolist()
        multi_channels2_list = self.multichannel_config['s2_offset'].tolist()
        multi_channeldet_list = self.multichannel_config['det_channel'].tolist()

        # Normalize detector counts by efficiency
        df = df.copy()  # Avoid modifying original data
        for tubeindex, tubename in enumerate(multi_detname_list):
            if tubename in df.columns:
                df[tubename] = df[tubename].astype(float)
                if multi_deteff_list[tubeindex] != 0:
                    df[tubename] = df[tubename] / multi_deteff_list[tubeindex]

        # Process each channel
        for channelindex, s2_offset in enumerate(multi_channels2_list):
            df[f"channel_s2_{channelindex}"] = df['s2'] + s2_offset
            df[f"channel_det_{channelindex}"] = 0  # Initialize to zero
            
            # Sum detector counts for this channel
            for tubenum in multi_channeldet_list[channelindex]:
                tube_col = f"tube{int(tubenum)}"
                if tube_col in df.columns:
                    df[f"channel_det_{channelindex}"] += df[tube_col]
            
            # Calculate HKL for each data point
            newhklarray = np.zeros((len(df), 3))
            for i in range(len(df)):
                try:
                    newhkl = AnglesToQhkl(
                        df["m2"].iloc[i],
                        df["s1"].iloc[i],
                        df[f"channel_s2_{channelindex}"].iloc[i],
                        df["a2"].iloc[i],
                        df["sgu"].iloc[i],
                        df["sgl"].iloc[i],
                        UBMat
                    )
                    newhklarray[i] = newhkl[:3].flatten()
                except (IndexError, KeyError, ValueError) as e:
                    print(f"Warning: Error calculating HKL for point {i}: {e}")
                    continue

            df[f"channel_h_{channelindex}"] = newhklarray[:, 0]
            df[f"channel_k_{channelindex}"] = newhklarray[:, 1]
            df[f"channel_l_{channelindex}"] = newhklarray[:, 2]

            # Create channel-specific DataFrame
            required_cols = [f"channel_h_{channelindex}", f"channel_k_{channelindex}", 
                           f"channel_l_{channelindex}", 'en', 
                           f"channel_det_{channelindex}", 'monitor']
            
            if all(col in df.columns for col in required_cols):
                df_channel = df[required_cols].copy()
                df_channel.columns = ['qh', 'qk', 'ql', 'en', 'detector', 'monitor']
                df_channel.attrs = df.attrs.copy()
                df_channel.attrs["channel"] = f"channel#{channelindex}"
                multichannel_dflist.append(df_channel)

        return multichannel_dflist

    def tas_multichannel_plot3d(self, dflist=None, xcol='qh', ycol='ql', zcol="en", 
                               perc_low=5, perc_high=95, xlim=None, ylim=None, zlim=None,
                               xlabel="QH [r.l.u]", ylabel="QL [r.l.u]", zlabel="E [meV]"):
        """
        Create 3D scatter plot of multichannel data.
        
        Args:
            dflist (list): List of DataFrames
            xcol (str): X-axis column name
            ycol (str): Y-axis column name
            zcol (str): Z-axis column name
            perc_low (float): Lower percentile for color scaling
            perc_high (float): Upper percentile for color scaling
            xlim (list): X-axis limits
            ylim (list): Y-axis limits
            zlim (list): Z-axis limits
            xlabel (str): X-axis label
            ylabel (str): Y-axis label
            zlabel (str): Z-axis label
            
        Returns:
            tuple: (X, Y, Z, A, ax) - coordinates, intensities, and axes
        """
        if not dflist:
            raise ValueError("dflist cannot be empty")
            
        required_cols = [xcol, ycol, zcol, 'detector']
        for df in dflist:
            if not all(col in df.columns for col in required_cols): 
                raise ValueError(f"Missing required columns: {required_cols}")

        X = np.array([])
        Y = np.array([])
        Z = np.array([]) 
        A = np.array([])

        for df in dflist:
            X = np.append(X, df[xcol].to_numpy())
            Y = np.append(Y, df[ycol].to_numpy())
            Z = np.append(Z, df[zcol].to_numpy())
            A = np.append(A, df['detector'].to_numpy())

        # Handle zero or negative values for log scale
        logA = np.log10(np.maximum(A, 1e-5))
        vminA, vmaxA = np.percentile(logA, [perc_low, perc_high])
        norm = Normalize(vmin=vminA, vmax=vmaxA, clip=True)
        scaled = norm(logA)
        colors = cm.viridis(scaled)
        colors[:, -1] = scaled  # Set alpha channel

        fig = plt.figure(figsize=(8, 6))
        ax = fig.add_subplot(111, projection='3d')
        sc = ax.scatter(X, Y, Z, c=colors)
        
        # Set axis limits if provided
        if xlim is not None and len(xlim) == 2:
            ax.set_xlim(xlim[0], xlim[1])
        if ylim is not None and len(ylim) == 2:
            ax.set_ylim(ylim[0], ylim[1])
        if zlim is not None and len(zlim) == 2:
            ax.set_zlim(zlim[0], zlim[1])

        # Labels and formatting
        ax.set_xlabel(xlabel, fontsize=14, labelpad=10)
        ax.set_ylabel(ylabel, fontsize=14, labelpad=10)
        ax.set_zlabel(zlabel, fontsize=14, labelpad=10)
        
        # Tick formatting
        ax.tick_params(axis='both', which='major', labelsize=12, width=2, length=6)
        ax.tick_params(axis='both', which='minor', width=1.5, length=4)
        
        plt.tight_layout()
        plt.show()
        
        return X, Y, Z, A, ax